# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details
# http://www.gnu.org/licenses/gpl-3.0.txt

"""Load urwid palette from file

This module handles loading urwid palettes from files or lists of strings and
applying them to a Screen object (see `ScreenBase.register_palette`).

First, load the default palette you ship with your application.

>>> init('path/to/default.palette', screen)

Now users can load their own palettes that overload default attributes.

>>> load('path/to/user.palette', screen)

Instead of a file you can also provide an iterable of strings or any object
with a 'readlines' method that returns such an iterable.

Use `reset` to go back to the defaults.

>>> reset(screen)


FILE FORMAT
===========

The palette format is very simple. Each line is either a palette entry or a
variable declaration.

Palette entries look like this:

    <ATTRIBUTE NAME> <FOREGROUND> on <BACKGROUND>

ATTRIBUTE NAME is a string without spaces. FOREGROUND and BACKGROUND are
16-color and 256-color attributes as specified in the urwid manual.

Variables are declared like that:

    $<NAME> = <VALUE>

Variables can be used as FOREGROUND or BACKGROUND attributes.

    $bg = dark gray
    this-attribute yellow     on $bg
    that-attribute light blue on $bg

"""

from collections import abc
import re
from urwid import (AttrSpec, AttrSpecError)


class ThemeError(Exception): pass
class ParseError(ThemeError): pass
class ValidationError(ThemeError): pass


def init(palette, screen):
    """Read `palette`, set it as default and apply it to `screen`

    palette: Where to get the palette from:
             - any object with a `readlines` method (e.g. a file handle),
             - a file name,
             - an iterable of lines.
    screen: A urwid screen instance (e.g. `raw_display.Screen`)

    The palette is passed to `set_default` before `register`ing it.
    """
    lines = read(palette)
    palette = Palette(lines)
    set_default(palette)
    register(screen, palette)

def load(palette, screen):
    """Same as `init`, but with validation

    The palette is passed to `validate` before `register`ing it.
    """
    lines = read(palette)
    palette = Palette(lines)
    validate(palette)
    register(screen, palette)


def read(source):
    """Read palette from `source`

    source: file handle, file name or iterable of strings

    Returns list of string lines.
    Raises TypeError or OSError.
    """
    if hasattr(source, 'read') and callable(source.readlines):
        lines = source.readlines()
    elif isinstance(source, str):
        try:
            with open(source, mode='r') as f:
                lines = f.readlines()
        except OSError as e:
            raise ThemeError('Unable to read {!r}: {}'.format(source, e.strerror))
    elif isinstance(source, abc.Iterable):
        lines = source
    else:
        raise TypeError('Invalid source: {!r}'.format(source))
    return [l.rstrip('\n') for l in lines]


def register(screen, palette):
    """Register palette with screen"""
    screen.set_terminal_properties(colors=palette.colors,
                                   bright_is_bold=palette.light_is_bold)
    screen.register_palette(palette)
    screen.clear()


DEFAULT_PALETTE = None
def set_default(palette):
    """Store copy of palette in global variable `DEFAULT_PALETTE`"""
    if not isinstance(palette, Palette):
        raise TypeError('Not a Palette: {!r}'.format(palette))
    global DEFAULT_PALETTE
    DEFAULT_PALETTE = palette.copy()


def validate(palette):
    """Validate palette (`set_default` must be called before)

    Raise ValidationError if any attribute name in `palette` is not in
    `DEFAULT_NAMES`.
    """
    if DEFAULT_PALETTE is None:
        raise RuntimeError('No defaults have been specified with set_default()')
    else:
        for entry in palette:
            name = entry[0]
            if name not in DEFAULT_PALETTE.names:
                raise ValidationError('Invalid attribute name: {!r}'.format(name))

def reset(screen):
    """`register` palette previously given to `set_default`"""
    if DEFAULT_PALETTE is None:
        raise RuntimeError('No defaults have been specified with set_default()')
    else:
        register(screen, DEFAULT_PALETTE)


class Palette(abc.Sequence):
    """Read urwid palette as a list of strings

    Raises ParseError on errors during parsing (duh).
    """

    VAR_REGEX = re.compile(r'^\$(?P<var>\S+) *= *(?P<val>.*?) *$')
    ATTR_REGEX = re.compile(r'^(?P<name>\S+) *(?P<fg>.*?) +on +(?P<bg>.*) *$')

    def __init__(self, lines):
        def apply_variables(string, variables):
            for name,value in sorted(variables.items(), key=lambda k: len(k[0]), reverse=True):
                string = string.replace('$'+name, value)
            return string

        def check_color(string):
            try:
                attrspec = AttrSpec(string, 'default')
            except AttrSpecError:
                raise ValueError(string)
            else:
                return max(16, attrspec.colors)

        def make_entry(name, fg, bg, colors):
            if colors == 256:
                return (name, 'default', 'default', 'default', fg, bg)
            elif colors == 16:
                return (name, fg, bg)
            else:
                raise RuntimeError('Unsupported number of colors: {}'.format(colors))

        self._light_is_bold = False
        palette = []
        variables = {}
        max_colors = 1
        for linenum,line in enumerate(lines, start=1):
            line = line.strip()
            if len(line) < 1 or line[0] == '#':
                # Ignore comments and empty lines
                continue

            if line == 'light_is_bold':
                self.light_is_bold = True
                continue

            # Read variable declaration
            match = self.VAR_REGEX.match(line)
            if match is not None:
                name, value = match.group('var', 'val')
                variables[name] = value
                continue

            # Read palette entry
            match = self.ATTR_REGEX.match(line)
            if match is not None:
                name, fg, bg = match.group('name', 'fg', 'bg')
                fg = apply_variables(fg, variables)
                bg = apply_variables(bg, variables)
                try:
                    this_colors = max(check_color(fg), check_color(bg))
                except ValueError as e:
                    raise ParseError('Invalid color in line {}: {!r}'.format(linenum, str(e)))
                else:
                    max_colors = max(max_colors, this_colors)
                    palette.append(make_entry(name, fg, bg, this_colors))
                    continue

            raise ParseError('Invalid line #{}: {!r}'.format(linenum, line))

        self._palette = palette
        self._colors = max_colors
        self._names = tuple(entry[0] for entry in palette)

    @property
    def light_is_bold(self):
        """Whether light colors are automatically bold"""
        return self._light_is_bold

    @light_is_bold.setter
    def light_is_bold(self, light_is_bold):
        self._light_is_bold = light_is_bold

    @property
    def colors(self):
        """Number of colors needed"""
        return self._colors

    @property
    def names(self):
        """Yield attribute names"""
        for name in self._names:
            yield name

    def copy(self):
        palette = Palette([])
        palette._palette = self._palette
        palette._colors = self._colors
        palette._names = self._names
        return palette

    def __eq__(self, other):
        return tuple(self) == tuple(other)

    def __ne__(self, other):
        return tuple(self) != tuple(other)

    def __repr__(self):
        return repr(self._palette)

    def __getitem__(self, item):
        return self._palette[item]

    def __iter__(self):
        return iter(self._palette)

    def __len__(self):
        return len(self._palette)



# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details
# http://www.gnu.org/licenses/gpl-3.0.txt

from ..logging import make_logger
log = make_logger(__name__)

import urwid
from . import urwidpatches

from ..main import (aioloop, cfg, cmdmgr, srvapi, helpmgr)
from . import theme

from ..settings.defaults import DEFAULT_KEYMAP
from .keymap import KeyMap
keymap = KeyMap(callback=lambda cmd,widget: cmdmgr.run_task(cmd, on_error=log.error))
for args in DEFAULT_KEYMAP:
    if args['action'][0] == '<' and args['action'][-1] == '>':
        args['action'] = keymap.mkkey(args['action'])
    keymap.bind(**args)
helpmgr.keymap = keymap


from .group import Group
from .tabs import Tabs
from .cli import CLIEditWidget
from .logger import LogWidget
from .infobar import (KeyChainsWidget, QuickHelpWidget, ConnectionStatusWidget,
                      BandwidthStatusWidget, TorrentCountersWidget)


def load_theme(themefile):
    """Load theme from `themefile`

    If `themefile` does not have a path and does not exist in the current
    working directory, try to load it from the same path as the rc file.
    """
    import os
    if os.sep not in themefile and not os.path.exists(themefile):
        # Path is not given and file does not exist in working dir.
        from ..settings.defaults import DEFAULT_RCFILE
        from ..main import cliargs
        rcfilepath = cliargs['rcfile'] or DEFAULT_RCFILE
        themefilepath = os.path.join(os.path.dirname(rcfilepath), themefile)
        if os.path.exists(themefilepath):
            theme.load(themefilepath, urwidscreen)
            return
    theme.load(themefile, urwidscreen)


def _create_cli_widget():
    def on_cancel(widget):
        widget.set_edit_text('')
        widgets.hide('cli')

    def on_accept(widget):
        cmd = widget.get_edit_text()
        on_cancel(widget)
        cmdmgr.run_task(cmd, on_error=log.error)

    return CLIEditWidget(':',
                         on_accept=on_accept, on_cancel=on_cancel,
                         history_file=cfg['tui.cli.history'].value)

def _greedy_spacer():
    return urwid.Padding(urwid.Text(''))

topbar = Group(cls=urwid.Columns)
topbar.add(name='host',   widget=ConnectionStatusWidget(), options='pack')
topbar.add(name='spacer', widget=urwid.AttrMap(_greedy_spacer(), 'topbar'))
topbar.add(name='help',   widget=QuickHelpWidget(), options='pack')

tabs = keymap.wrap(Tabs, context='tabs')(
    tabbar=urwid.AttrMap(urwid.Columns([], dividechars=2), 'tabs.unfocused')
)

bottombar = Group(cls=urwid.Columns)
bottombar.add(name='counters', widget=TorrentCountersWidget(), options='pack')
bottombar.add(name='spacer', widget=urwid.AttrMap(_greedy_spacer(), 'bottombar'))
bottombar.add(name='bandwidth', widget=BandwidthStatusWidget(), options='pack')

cli = urwid.AttrMap(_create_cli_widget(), 'cli')

logwidget = LogWidget(maxrows=cfg['tui.log.height'].value,
                      autohide_delay=cfg['tui.log.autohide'].value)

keychains = KeyChainsWidget()
keymap.on_keychain(keychains.update)

widgets = keymap.wrap(Group, context='main')(cls=urwid.Pile)
widgets.add(name='topbar', widget=topbar, options='pack')
widgets.add(name='main', widget=tabs)
widgets.add(name='log', widget=logwidget, options='pack')
widgets.add(name='cli', widget=cli, options='pack', visible=False)
widgets.add(name='bottombar', widget=bottombar, options='pack')
widgets.add(name='keychains', widget=keychains, options='pack')


def unhandled_input(key):
    key = keymap.evaluate(key)
    if key is not None:
        log.debug('Unhandled key: %s', key)

urwidscreen = urwid.raw_display.Screen()
urwidloop = urwid.MainLoop(widgets,
                           screen=urwidscreen,
                           event_loop=urwid.AsyncioEventLoop(loop=aioloop),
                           unhandled_input=unhandled_input,
                           handle_mouse=False)


def run(command_runner):
    """Run commands and start TUI

    Return False if any of the commands failed, True otherwise.
    """
    # Don't catch theme.ParserError - a broken default theme should make us
    # croak obviously and horribly
    theme.init(cfg['tui.theme'].default, urwidscreen)

    # Load tui-specific hooks before commands run (commands may trigger hooks)
    from . import hooks

    try:
        if not command_runner():
            return False
    # Make 'quit' behave as expected
    except urwid.ExitMainLoop:
        return True

    # Start logging to TUI widget instead of stdout/stderr
    logwidget.enable()

    # If no tabs have been opened by cli or rc file, open default tab
    if len(tabs) <= 0:
        cmdmgr.run_sync('ls', on_error=log.error)

    try:
        # Start polling torrent lists, counters, bandwidth usage, etc.
        aioloop.run_until_complete(srvapi.start_polling())
        urwidloop.run()
    finally:
        logwidget.disable()
        aioloop.run_until_complete(srvapi.stop_polling())

    return True



# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details
# http://www.gnu.org/licenses/gpl-3.0.txt

"""Torrent commands for the TUI"""

from ...logging import make_logger
log = make_logger(__name__)

from collections import abc

from ..base import torrent as base
from . import mixin
from .. import ExpectedResource
from ...utils import strcrop
from . import make_tab_title


class AddTorrentsCmd(base.AddTorrentsCmdbase,
                     mixin.polling_frenzy, mixin.make_request):
    provides = {'tui'}


class AnnounceTorrentsCmd(base.AnnounceTorrentsCmdbase,
                          mixin.make_request, mixin.select_torrents):
    provides = {'tui'}


class ListTorrentsCmd(base.ListTorrentsCmdbase,
                      mixin.select_torrents):
    provides = {'tui'}
    tui = ExpectedResource

    def make_tlist(self, tfilter, sort, columns):
        # Enforce 'marked' column
        if 'marked' not in columns:
            columns.insert(0, 'marked')

        def make_title_widget(text):
            return make_tab_title(text,
                                  'tabs.torrentlist.unfocused',
                                  'tabs.torrentlist.focused')

        from ...tui.torrent.tlist import TorrentListWidget
        tlistw = TorrentListWidget(tfilter=tfilter, sort=sort, columns=columns)
        tabid = self.tui.tabs.load(make_title_widget(tlistw.title), tlistw)

        def set_tab_title(text):
            self.tui.tabs.set_title(make_title_widget(text), position=tabid)
        tlistw.title_updater = set_tab_title

        return True


class ListFilesCmd(base.ListFilesCmdbase,
                   mixin.make_request, mixin.select_torrents, mixin.select_files):
    provides = {'tui'}
    tui = ExpectedResource
    srvapi = ExpectedResource

    async def make_flist(self, tfilter, ffilter, columns):
        if isinstance(tfilter, abc.Sequence) and len(tfilter) == 1:
            # tfilter is a torrent ID - resolve it to a name for the title
            response = await self.srvapi.torrent.torrents(tfilter, keys=('name',))
            title = strcrop(response.torrents[0]['name'], 30, tail='…')
        else:
            title = None

        # Enforce 'marked' column
        if 'marked' not in columns:
            columns.insert(0, 'marked')

        from ...tui.torrent.flist import FileListWidget
        flistw = FileListWidget(self.srvapi, tfilter, ffilter, columns, title=title)

        def make_title_widget(text):
            return make_tab_title(text,
                                  'tabs.filelist.unfocused',
                                  'tabs.filelist.focused')

        tabid = self.tui.tabs.load(make_title_widget(flistw.title), flistw)

        def set_tab_title(text):
            self.tui.tabs.set_title(make_title_widget(text), position=tabid)
        flistw.title_updater = set_tab_title

        return True


class ListPeersCmd(base.ListPeersCmdbase,
                   mixin.make_request, mixin.select_torrents):
    provides = {'tui'}
    tui = ExpectedResource
    srvapi = ExpectedResource

    async def make_plist(self, tfilter, pfilter, sort, columns):
        if isinstance(tfilter, abc.Sequence) and len(tfilter) == 1:
            # tfilter is a torrent ID - resolve it to a name for the title
            response = await self.srvapi.torrent.torrents(tfilter, keys=('name',))
            title = strcrop(response.torrents[0]['name'], 30, tail='…')
        else:
            title = None

        from ...tui.torrent.plist import PeerListWidget
        plistw = PeerListWidget(self.srvapi, tfilter=tfilter, pfilter=pfilter,
                                sort=sort, columns=columns, title=title)

        def make_title_widget(text):
            return make_tab_title(text,
                                  'tabs.peerlist.unfocused',
                                  'tabs.peerlist.focused')

        tabid = self.tui.tabs.load(make_title_widget(plistw.title), plistw)

        def set_tab_title(text):
            self.tui.tabs.set_title(make_title_widget(text), position=tabid)
        plistw.title_updater = set_tab_title

        return True


class MoveTorrentsCmd(base.MoveTorrentsCmdbase,
                      mixin.polling_frenzy, mixin.make_request, mixin.select_torrents):
    provides = {'tui'}


class PriorityCmd(base.PriorityCmdbase,
                  mixin.polling_frenzy, mixin.make_request, mixin.select_torrents, mixin.select_files):
    provides = {'tui'}


class RemoveTorrentsCmd(base.RemoveTorrentsCmdbase,
                        mixin.polling_frenzy, mixin.make_request, mixin.select_torrents):
    provides = {'tui'}


class StartTorrentsCmd(base.StartTorrentsCmdbase,
                       mixin.polling_frenzy, mixin.make_request, mixin.select_torrents):
    provides = {'tui'}


class StopTorrentsCmd(base.StopTorrentsCmdbase,
                      mixin.polling_frenzy, mixin.make_request, mixin.select_torrents):
    provides = {'tui'}


class VerifyTorrentsCmd(base.VerifyTorrentsCmdbase,
                        mixin.make_request, mixin.select_torrents):
    provides = {'tui'}



# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details
# http://www.gnu.org/licenses/gpl-3.0.txt

"""Create and manage commands

Commands can be created anywhere by defining a class with metaclass set to
InitCommand.

>>> class MyCommand(metaclass=InitComand):
>>>     ...

Every command class must have a 'run' method, which is called when the command
is executed. It can be a coroutine function or a normal, synchronous function.
The command's arguments are passed to 'run' as keywords.

The 'run' method should return True on success and False on failure.  It may
also raise CmdError.  If False is returned, it is assumed that any errors have
already been reported (i.e. logged to the ERROR level).

If a command executes multiple actions (e.g. pausing two torrents) and any of
those actions fails, it should be a failure (e.g. first torrent was paused,
second torrent does not exist -> return False).

Every command class must have the following class attributes:

    run (callable):            Method that is called when the command runs
    name (string):             Name of the command
    category (string):         Name of the category (used only for `help commands`)
    provides (set of strings): Supported interfaces ('cli' and/or 'tui')
    description (string):      One-line description for command

These class attributes are optional:

    aliases (sequence of strings):   Alternative names
    argspecs (sequence of mappings): argparse specs for arguments
    usage (sequence of strings):     Syntax string of any arguments (in man pages
                                     this is called SYNOPSIS)
    examples (sequence of strings):  Typical invocations
    more_sections (dict):            Maps additional section names to list of strings or
                                     callables that return lists of strings

Arguments
---------

Each command class gets its own argparser.ArgumentParser instance.  Command
arguments are specified by setting the 'argspecs' class attribute to a
sequence of mappings.  The mappings are then used to create arguments for
ArgumentParser.add_argument.  Most items are passed on unmodified, but there
are exceptions:

- 'names' must be set to a string or a sequence of strings that is passed on
  as the positional arguments to add_argument.  The first string is used as
  the keyword when providing the argument to the run method.

- 'description' must be set to a string that describes what the command does.

- 'nargs' can be the string 'REMAINDER', which is replaced with the value of
  argparse.REMAINDER.

Resources
---------

Resources are objects that are provided to the CommandManager instance as
key-value pairs which sets them as attributes for command classes that want
them.  For example, the global client API instance 'srvapi' is a resource.

Command classes can request resources by setting a class attribute to
ExpectedResource.  If the CommandManager instance has a resource by this name,
it will be replaced by InitCommand and commands can use it like any regular
class attribute.  If the resource does not exist, accessing it raises an
AttributeError.
"""

from ..logging import make_logger
log = make_logger(__name__)

import asyncio
import argparse
import shlex
from inspect import getmembers
from importlib import import_module
from collections import abc
from .utils import CallbackDict


def ExpectedResource(name):
    def _raise(self):
        raise AttributeError('{} misses expected resource: {}'.format(type(self).__name__, name))
    return property(fget=_raise)


OPS_AND = ('&', 'and')
OPS_OR  = ('|', 'or')
OPS_SEQ = (';', 'also')

def is_op(string):
    return any(string in ops for ops in (OPS_SEQ, OPS_AND, OPS_OR))


class CmdError(Exception): pass
class CmdArgError(CmdError): pass
class CmdNotFoundError(CmdError): pass


class StayAliveArgParser(argparse.ArgumentParser):
    """Capture errors instead of printing them on stderr and exiting"""
    def exit(self, status=0, message=None):
        if message is not None:
            self.error(message)

    def error(self, message):
        raise CmdArgError(message[0].upper() + message[1:])


def iscmdcls(obj):
    """Whether `obj` is a command class"""
    return isinstance(obj, type) and \
        obj is not _CommandBase and \
        issubclass(obj, _CommandBase)


_MANDATORY_CMD_ATTRS = ('name', 'category', 'provides', 'description', 'run')
_OPTIONAL_CMD_ATTRS = {
    'aliases': (), 'usage': (), 'examples': (), 'argspecs': (), 'more_sections': {}
}
def InitCommand(clsname, bases, attrs):
    """Class factory that inits all commands

    This takes care of inheritance, complains about missing mandatory class
    attributes, adds defaults for optional class attributes, etc.
    """
    # Make sure all mandatory attributes are there
    for attr in _MANDATORY_CMD_ATTRS:
        if attr not in attrs:
            raise RuntimeError('{} misses mandatory attribute: {!r}'
                               .format(clsname, attr))

    # Add defaults for optional attributes
    for k,v in _OPTIONAL_CMD_ATTRS.items():
        if k not in attrs:
            attrs[k] = v

    # Turn ExpectedResource into a property that raises an exception if not
    # provided by CommandManager
    for k,v in attrs.items():
        if v is ExpectedResource:
            attrs[k] = ExpectedResource(k)

    # 'names' attribute
    attrs['names'] = list(attrs['aliases'])
    attrs['names'].insert(0, attrs['name'])

    # Create argument parser
    argp = StayAliveArgParser()
    for argspec in attrs['argspecs']:
        # Check if all mandatory keys are in the spec
        for mandatory_key in ('names', 'description'):
            if mandatory_key not in argspec:
                raise RuntimeError('Missing key {!r} in argument spec: {}'
                                   .format(mandatory_key, argspec))

        # Create a copy of the spec for this arg and remove all items that
        # ArgParser doesn't understand
        argspec = argspec.copy()
        argspec.pop('description')
        if 'default_description' in argspec:
            argspec.pop('default_description')

        # Assemble arg names
        argnames = argspec.pop('names', None)
        if not isinstance(argnames, abc.Sequence):
            argnames = (argnames,)

        # Translate string to argparse constant
        if 'nargs' in argspec and argspec['nargs'] == 'REMAINDER':
            argspec['nargs'] = argparse.REMAINDER

        # Create new argparser
        argp.add_argument(*argnames, **argspec)
    attrs['_argparser'] = argp

    # Command class must inherit from _CommandBase
    if _CommandBase not in bases:
        bases = (_CommandBase,) + bases

    cls = type(clsname, bases, attrs)
    return cls


class _CommandBase():
    """Base for all commands classes"""
    def __init__(self, args=(), on_success=None, on_error=None, loop=None, **kwargs):
        for k,v in kwargs.items():
            setattr(self, k, v)
        self._loop = loop
        self._args = args
        self.on_success = on_success
        self.on_error = on_error
        self._task = None
        self._success = None
        self._is_async = False
        self._exc = None
        self._exc_fetched = False

        try:
            args_parsed = self._argparser.parse_args(args)
        except CmdArgError as e:
            self._finish(success=False,
                         exception=CmdArgError('{}: {}'.format(self.name, e)))
        else:
            # Create keyword args for run() method
            kwargs = {}
            for argspec in self.argspecs:
                # First name is the kwarg for run()
                key = argspec['names'][0].lstrip('-').replace('-', '_')
                value = getattr(args_parsed, key)
                kwargs[key.replace(' ', '_')] = value
            self._args = kwargs

            if asyncio.iscoroutinefunction(self.run):
                log.debug('Running async command: %r', self)
                self._is_async = True
                self._task = self.loop.create_task(self.run(**kwargs))
                self._task.add_done_callback(lambda task: self._catch_exceptions(task.result))
            else:
                log.debug('Running sync command: %r', self)
                self._is_async = False
                self._catch_exceptions(self.run, **kwargs)

    def _catch_exceptions(self, callabee, *args, **kwargs):
        try:
            success = callabee(*args, **kwargs)
        except Exception as e:
            success = False
            exc = e
        else:
            exc = None
        self._handle_result(success, exc)

    def _handle_result(self, success, exception):
        if success:
            self._finish(success=True)
        else:
            if exception is None:
                exception = CmdError()
            self._finish(success=False, exception=exception)

    def _finish(self, success, exception=None):
        if self.finished:
            return  # _finish() was already called

        self._success = success
        self._exc = exception

        if success:
            if self.on_success is not None:
                log.debug('Calling success callback: %r', self.on_success)
                self.on_success(self)
        else:
            if self.on_error is not None:
                log.debug('Calling error callback: %r', self.on_error)
                self.on_error(self.exception)

    def wait_sync(self):
        """Wait synchronously until this command has finished

        This uses the run_until_complete() method, so the loop must not be
        running or closed.
        """
        if not self.finished:
            if self.loop.is_closed():
                raise RuntimeError('AsyncIO loop is closed - cannot wait for {!r}'.format(self))
            elif self.loop.is_running():
                raise RuntimeError('AsyncIO loop is running - cannot wait for {!r}'.format(self))
            else:
                log.debug('Waiting until finished: %r', self)
                self._catch_exceptions(self.loop.run_until_complete, self._task)

    async def wait_async(self):
        """Wait asynchronously until this command has finished"""
        if not self.finished:
            log.debug('Waiting until finished: %r', self)
            try:
                success = await self._task
            except Exception as e:
                self._handle_result(success=False, exception=e)
            else:
                self._handle_result(success=success, exception=None)

    def __del__(self):
        """Raise stored, unraised exception"""
        exc = self.exception
        if exc and not self._exc_fetched:
            raise exc

    @property
    def loop(self):
        """AsyncIO loop"""
        if self._loop is None:
            raise RuntimeError('Called with missing asyncio loop: {!r}'.format(self))
        else:
            return self._loop

    @property
    def success(self):
        """True/False if command has finished, None otherwise"""
        return self._success

    @property
    def exception(self):
        """Exception if command raised one, None otherwise"""
        # If this class is called with wrong arguments, __del__ is called
        # before _exc exists.
        if hasattr(self, '_exc'):
            self._exc_fetched = True
            return self._exc

    @property
    def finished(self):
        """Whether command has finished"""
        return self._success is not None

    @property
    def is_async(self):
        """Whether this command runs asynchronously"""
        return self._is_async

    def __repr__(self):
        if isinstance(self._args, abc.Sequence):
            argstr = ', '.join(repr(arg) for arg in self._args)
        elif isinstance(self._args, abc.Mapping):
            argstr = ', '.join('%s=%r' % (k,v)
                                for k,v in self._args.items())
        provides = '/'.join(interface for interface in self.provides)
        string = '<Command [{}] {}({})'.format(provides, self.name, argstr)
        if self.finished:
            string += ' success={}'.format(self.success)
        else:
            string += ' running'
        return string + '>'


class CommandManager():
    def __init__(self, loop=None):
        self.loop = loop if loop is not None else asyncio.get_event_loop()
        self._cmds = {}
        self._active_interface = None
        self._resources = CallbackDict(callback=self._update_resources)
        from . import utils
        self._resources.update(cmdmgr=self)

    def load_cmds_from_module(self, *modules):
        """Import modules and look for command classes

        Command classes must have set metaclass=InitCommand.
        """
        for modname in modules:
            log.debug('Looking for commands in %s', modname)
            mod = import_module(modname)
            for member in getmembers(mod):
                cmdcls = member[1]
                # We're interested in classes that inherit from _CommandBase
                # but are not _CommandBase.
                if iscmdcls(cmdcls):
                    self.register(cmdcls)

    def register(self, cmdcls):
        """Add new command

        cmdcls: command class (must have InitCommand as metaclass)
        """
        if not issubclass(cmdcls, _CommandBase):
            raise RuntimeError("{} must have InitCommand as metaclass".format(cmdcls))
        elif not cmdcls.provides:
            raise RuntimeError('{} does not provide any interfaces'.format(cmdcls))

        self._provide_resources_to_cmdcls(cmdcls)

        # Store cmdcls for each of its supported interfaces
        for interface in cmdcls.provides:
            if interface not in self._cmds:
                self._cmds[interface] = {}
            log.debug('Registered %s command %s (%s)',
                      interface, cmdcls.name, type(cmdcls).__name__)
            self._cmds[interface][cmdcls.name] = cmdcls

    @property
    def resources(self):
        return self._resources

    def _update_resources(self):
        for cmdcls in self.all_commands:
            self._provide_resources_to_cmdcls(cmdcls)

    def _provide_resources_to_cmdcls(self, cmdcls):
        for name,obj in self._resources.items():
            if hasattr(cmdcls, name):
                setattr(cmdcls, name, obj)

    @property
    def active_interface(self):
        return self._active_interface

    @active_interface.setter
    def active_interface(self, interface):
        if interface in self._cmds:
            self._active_interface = interface
        else:
            raise ValueError('No commands for interface {!r} registered'.format(interface))

    @property
    def active_commands(self):
        """Tuple of commands for the active interface or all commands with no active interface"""
        if self._active_interface is None:
            return self.all_commands
        else:
            return tuple(self._cmds[self._active_interface].values())

    @property
    def all_commands(self):
        """Tuple of all commands for all interfaces"""
        cmds = set()
        for interface,cmdnames in self._cmds.items():
            for cmdname in cmdnames:
                cmds.add(self._cmds[interface][cmdname])
        return tuple(cmds)

    def get_cmdcls(self, cmdname, interface='ACTIVE', exclusive=False):
        """Resolve command name `cmdname` to command class

        If `interface` is 'ACTIVE', return command class from
        `active_commands`.

        If `interface` is 'ANY', return command class from `all_commands`.

        If `interface` is anything else, it must be an existing interface and
        only a command that supports it is returned.

        If `exclusive` evaluates to True, the returned command class does not
        support any other interfaces.

        Returns None if not matching command class is registered.
        """
        if interface == 'ACTIVE':
            cmdpool = self.active_commands
        elif interface == 'ANY':
            cmdpool = self.all_commands
        elif isinstance(interface, abc.Hashable):
            try:
                cmdpool = tuple(self._cmds[interface].values())
            except KeyError:
                raise ValueError('Unknown interface: {!r}'.format(interface))
        else:
            raise RuntimeError('Interface type must be hashable: {!r}'.format(interface))

        for cmd in cmdpool:
            if cmdname in cmd.names:
                if not exclusive:
                    return cmd
                elif cmd.provides == (interface,):
                    return cmd

    def __getitem__(self, cmdname):
        cmd = self.get_cmdcls(cmdname, interface='ANY')
        if cmd is not None:
            return cmd
        else:
            raise KeyError(cmdname)

    def __contains__(self, cmdname):
        return self.get_cmdcls(cmdname, interface='ANY') is not None

    @property
    def categories(self):
        """Tuple of command categories built from commands' `category` attributes"""
        categories = set()
        for cmd in self.all_commands:
            categories.add(cmd.category)
        return tuple(sorted(categories))


    def _maybe_run_callbacks(self, process, on_success, on_error):
        exc = process.exception
        if process.success:
            if on_success is not None:
                on_success(process)
            return True
        elif on_error is not None and isinstance(exc, CmdError):
            # A command raises an empty CmdError if its run() method returned
            # False.  That should mean the run() method already reported any
            # errors, so we ignore that error here.
            if exc.args:
                log.debug('Calling %r with %r', on_error, exc)
                on_error(exc)
                return False
            else:
                log.debug('Ignoring empty CmdError')
                return False
        elif exc is not None:
            log.debug('Re-raising %r', exc)
            raise exc

    def _handle_final_process(self, process, reraise=True):
        log.debug('Final process: %r', process)
        if process is None:
            return True  # Empty cmd chain
        elif process.exception is not None:
            if reraise:
                raise process.exception
            else:
                return False
        else:
            return process.success

    def run_sync(self, commands, on_success=None, on_error=None, **kwargs):
        """Run `commands`, return boolean result

        Use this method in non-async code. The asyncio loop must be not
        running and not closed.

        command: See `split_cmdchain`
        block: Whether to return immediately or wait for an async command to
               finish (if enabled, the asyncio event loop must be stopped and
               not closed)
        on_error: Callable that gets any CmdErrors

        Any other keyword arguments are forwarded to the command instance
        which makes them available as attributes so the 'run' method can
        access them via its instance (self).

        Any exceptions from the command (besides CmdErrors) are re-raised
        unless.  If `on_error` is None, CmdErrors are also re-raised.

        Returns True if the command chain ran successfully, False if it
        failed, and None if it is async and `block` is False.

        """
        log.debug('Running command chain synchronously: %r', commands)
        cmdchain = self.split_cmdchain(commands)
        process = None
        for process in self._yield_from_cmdchain(cmdchain, **kwargs):
            if not process.finished:
                process.wait_sync()
            self._maybe_run_callbacks(process, on_success, on_error)
        return self._handle_final_process(process, reraise=not bool(on_error))

    async def run_async(self, commands, on_success=None, on_error=None, **kwargs):
        """Same as `run_sync` but for async code"""
        log.debug('Running command chain asynchronously: %r', commands)
        cmdchain = self.split_cmdchain(commands)
        process = None
        for process in self._yield_from_cmdchain(cmdchain, **kwargs):
            if not process.finished:
                await process.wait_async()
            self._maybe_run_callbacks(process, on_success, on_error)
        return self._handle_final_process(process, reraise=not bool(on_error))

    def run_task(self, commands, on_success=None, on_error=None, **kwargs):
        """Return task that runs `run_async`"""
        log.debug('Creating command chain task: %r', commands)
        return self.loop.create_task(self.run_async(
            commands, on_success, on_error, **kwargs))

    def _yield_from_cmdchain(self, cmdchain, **kwargs):
        prev_process_success = True
        prev_process_exc = None
        for i,cmdline in enumerate(cmdchain):
            if cmdline in OPS_AND and not prev_process_success:
                log.debug('Previous command failed (%r) - aborting', prev_process_success)
                break
            elif cmdline in OPS_OR and prev_process_success:
                log.debug('Previous command succeeded (%r) - aborting', prev_process_success)
                break
            elif is_op(cmdline):
                continue
            else:
                process = self._create_process(cmdline, **kwargs)
                yield process
                assert process.finished, 'Not finished: {!r}'.format(process)
                prev_process_success = process.success

    def _create_process(self, cmdline, **kwargs):
        """Call one command and return its instance or None on error"""
        # We can't create a _CommandBase instance for non-existing commands,
        # and raising/returning an exception instead of a _CommandBase
        # instance complicates everything.
        # This creates a rudimentary process with only the needed attributes
        # so the caller can report success or failure as if it were a proper
        # _CommandBase derivative.
        def fake_process(**kwargs):
            attrs = {'finished': True, 'success': False, 'exception': None}
            attrs.update(kwargs)
            def __repr__(self):
                kws = ', '.join('%s=%r' % (k,v) for k,v in attrs.items()
                                if k[0] != '_')
                return '<{} {}>'.format(type(self).__name__, kws)
            attrs['__repr__'] = __repr__
            return type('FakeCommand', (), attrs)()

        if not isinstance(cmdline, list):
            cmdline = list(cmdline)

        cmdname = cmdline.pop(0)
        cmdargs = cmdline
        cmdcls = self.get_cmdcls(cmdname, interface='ACTIVE')
        if cmdcls is None:
            if self.get_cmdcls(cmdname, interface='ANY') is not None:
                # Command exists but not in active interface - we ignore it
                log.debug('Ignoring inactive command: %r', cmdname)
                process = fake_process(success=None)
            else:
                exc = CmdNotFoundError('Unknown command: {}'.format(cmdname))
                process = fake_process(exception=exc)
        elif cmdcls not in self.active_commands:
            exc = CmdError('{}: No support for {} interface'
                           .format(cmdname, self._active_interface))
            process = fake_process(exception=exc)
        else:
            process = cmdcls(cmdargs, loop=self.loop, **kwargs)
        return process

    def split_cmdchain(self, commands):
        """Yield split commands and command operators

        commands: Command string or iterable

        Command strings must separate commands with command operators.

        Command iterables must yield command sequences (e.g. lists or tuples)
        and command chain operators.  The first item in a command sequence is
        the command's name and the rest is the arguments for the command.

        Command operators are characters specified by the variables OPS_AND,
        OPS_OR and OPS_SEQ surrounded by spaces (e.g. " & ").

        Simple example:
        >>> list(cmdmgr.split_cmdchain('do that & act "like this"'))
        [ ['do', 'that'], '&', ['act', 'like this'] ]
        """
        if isinstance(commands, str):
            args = shlex.split(commands)
            cmdchain = []
            cmd = []
            while args:
                arg = args.pop(0)
                if arg in OPS_SEQ:
                    if cmd:
                        cmdchain.append(cmd)
                    cmdchain.append(OPS_SEQ[0])
                    cmd = []
                elif arg in OPS_OR:
                    if cmd:
                        cmdchain.append(cmd)
                    cmdchain.append(OPS_OR[0])
                    cmd = []
                elif arg in OPS_AND:
                    if cmd:
                        cmdchain.append(cmd)
                    cmdchain.append(OPS_AND[0])
                    cmd = []
                else:
                    cmd.append(arg)
            if cmd:
                cmdchain.append(cmd)
            while cmdchain and is_op(cmdchain[-1]):
                cmdchain.pop(-1)
        elif not isinstance(commands, abc.Iterable):
            raise RuntimeError('Must be string or sequence, not {!r}: {!r}'.format(
                type(commands).__name__, commands))
        else:
            cmdchain = commands

        for item in cmdchain:
            self._validate_cmdchain_item(item)
            yield item

    def _validate_cmdchain_item(self, item):
        # Test if item is of a valid type
        if not (is_op(item) or (isinstance(item, abc.Sequence) and not isinstance(item, str) and
                                all(isinstance(arg, str) for arg in item))):
            raise RuntimeError('Invalid type for command chain item: {!r}'.format(item))

        # Test if item is an operator after another operator
        try:
            prev_item = self._prev_validation_item
        except AttributeError:
            prev_item = None
        self._prev_validation_item = item
        if is_op(prev_item) and is_op(item):
            raise CmdError('Consecutive operators: "{} {}"'.format(prev_item, item))



# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details
# http://www.gnu.org/licenses/gpl-3.0.txt

from ...logging import make_logger
log = make_logger(__name__)

from .. import tkeys as tkeys
from .. import utils
from .. import base

from os.path import normpath


# Some values need to be modified to comply with our internal standards

def _modify_ratio(raw_torrent):
    #define TR_RATIO_NA  -1
    #define TR_RATIO_INF -2
    ratio = raw_torrent['uploadRatio']
    return tkeys.Ratio.UNKNOWN if ratio in (-1, -2) else ratio

def _modify_eta(raw_torrent):
    #define TR_ETA_NOT_AVAIL -1
    #define TR_ETA_UNKNOWN -2
    seconds = raw_torrent['eta']
    if seconds == -1:
        return tkeys.Timedelta.NOT_APPLICABLE
    elif seconds == -2:
        return tkeys.Timedelta.UNKNOWN
    return seconds

def _count_seeds(raw_torrent):
    trackerStats = raw_torrent['trackerStats']
    if trackerStats:
        return max(t['seederCount'] for t in trackerStats)
    else:
        return tkeys.SeedCount.UNKNOWN

def _is_isolated(raw_torrent):
    """Return whether this torrent can find any peers via trackers or DHT"""
    if not raw_torrent['isPrivate']:
        return False  # DHT is used

    # Torrent has trackers?
    trackerStats = raw_torrent['trackerStats']
    if trackerStats:
        # Did we try to connect to a tracker?
        if any(tracker['hasAnnounced'] for tracker in trackerStats):
            # Did at least one tracker respond?
            if any(tracker['lastAnnounceSucceeded'] for tracker in trackerStats):
                return False
        # We didn't try yet; assume non-isolation
        else:
            return False
    return True  # No way to find any peers

_STATUS_MAP = {
    0: tkeys.Status.STOPPED,
    1: tkeys.Status.VERIFY_Q,
    2: tkeys.Status.VERIFY,
    3: tkeys.Status.LEECH_Q,
    4: tkeys.Status.LEECH,
    5: tkeys.Status.SEED_Q,
    6: tkeys.Status.SEED,
}



def _create_TorrentFileTree(raw_torrent):
    filelist = raw_torrent['fileStats']
    # Combine 'files' and 'fileStats' fields and add the 'id' field, which
    # is the index in the list provided by Transmission.
    if 'files' in raw_torrent:
        for i,(f1,f2) in enumerate(zip(filelist, raw_torrent['files'])):
            f1['id'] = i
            f1.update(f2)
    else:
        for i,f in enumerate(filelist):
            f['id'] = i
    return TorrentFileTree(raw_torrent['id'], entries=filelist)

import os
class TorrentFileTree(base.TorrentFileTreeBase):
    def __init__(self, torrent_id, entries, path=[]):
        if isinstance(entries, TorrentFileTree):
            self._items = entries._items
            return

        self._path = os.sep.join(path)
        items = {}
        subfolders = {}

        for entry in entries:
            parts = entry['name'].split(os.sep, 1)
            if len(parts) == 1:
                items[parts[0]] = tkeys.TorrentFile(
                    tid=torrent_id, id=entry['id'],
                    name=entry['name'], path=path,
                    size_total=entry['length'],
                    size_downloaded=entry['bytesCompleted'],
                    is_wanted=entry['wanted'],
                    priority=entry['priority'])

            elif len(parts) == 2:
                subfolder, subpath = parts
                if subfolder not in subfolders:
                    subfolders[subfolder] = []
                entry['name'] = subpath
                subfolders[subfolder].append(entry)
            else:
                raise RuntimeError(parts)

        for subfolder,entries in subfolders.items():
            items[subfolder] = TorrentFileTree(torrent_id, entries, path=path+[subfolder])
        self._items = items

    def update(self, fileStats):
        def update_files(ftree, fileStats):
            for entry in ftree.values():
                if isinstance(entry, tkeys.TorrentFile):
                    # File ID is its index in the list provided by
                    # Transmission (see _create_TorrentFileTree)
                    fstats = fileStats[entry['id']]
                    entry.update({'size-downloaded': fstats['bytesCompleted'],
                                  'is-wanted': fstats['wanted'],
                                  'priority': fstats['priority']})
                else:
                    update_files(entry, fileStats)

        update_files(self._items, fileStats)


class TrackerList(tuple):
    def __new__(cls, raw_torrent):
        return super().__new__(cls,
            ({'id': tracker['id'],
              'url-announce': utils.URL(tracker['announce'])}
             for tracker in raw_torrent['trackers'])
        )


class PeerList(tuple):
    def __new__(cls, raw_torrent):
        TorrentPeer = tkeys.TorrentPeer
        return super().__new__(cls,
            (TorrentPeer(tid=raw_torrent['id'], tname=raw_torrent['name'],
                         tsize=raw_torrent['totalSize'],
                         ip=p['address'], port=p['port'], client=p['clientName'],
                         progress=p['progress']*100,
                         rate_up=p['rateToPeer'], rate_down=p['rateToClient'])
             for p in raw_torrent['peers'])
        )


# Map our keys to tuples of needed RPC field names for those keys
DEPENDENCIES = {
    'id'                : ('id',),
    'hash'              : ('hashString',),
    'name'              : ('name',),
    'status'            : ('status',),
    'path'              : ('downloadDir',),

    'private'           : ('isPrivate',),
    'stalled'           : ('isStalled',),
    'isolated'          : ('trackerStats', 'isPrivate'),

    '%downloaded'       : ('percentDone',),
    '%metadata'         : ('metadataPercentComplete',),
    '%verified'         : ('recheckProgress',),

    'peers-connected'   : ('peersConnected',),
    'peers-uploading'   : ('peersSendingToUs',),
    'peers-downloading' : ('peersGettingFromUs',),
    'peers-seeding'     : ('trackerStats',),

    'timestamp-created' : ('dateCreated',),
    'timestamp-added'   : ('addedDate',),
    'timestamp-started' : ('startDate',),
    'timestamp-active'  : ('activityDate',),
    'timestamp-done'    : ('doneDate',),
    'timespan-eta'      : ('eta',),
    'timestamp-manual-announce-allowed': ('manualAnnounceTime',),

    'ratio'             : ('uploadRatio',),
    'rate-down'         : ('rateDownload',),
    'rate-up'           : ('rateUpload',),

    'size-final'        : ('sizeWhenDone',),
    'size-total'        : ('totalSize',),
    'size-downloaded'   : ('downloadedEver',),
    'size-uploaded'     : ('uploadedEver',),
    'size-available'    : ('desiredAvailable',),
    'size-corrupt'      : ('corruptEver',),

    'trackers'          : ('trackers',),
    'peers'             : ('peers', 'totalSize'),

    # 'files' is called once to initialize file names and sizes by
    # api_torrent.TorrentAPI._get_torrents_by_ids when files are requested.
    'files'             : ('fileStats',),
}

# Map our keys to callables that adjust the raw RPC values or create new
# values from existing RPC values.
_MODIFY = {
    '%downloaded'     : lambda raw: raw['percentDone']*100,
    '%metadata'       : lambda raw: raw['metadataPercentComplete']*100,
    '%verified'       : lambda raw: raw['recheckProgress']*100,
    'status'          : lambda raw: _STATUS_MAP[raw['status']],
    'peers-seeding'   : _count_seeds,
    'isolated'        : _is_isolated,
    'ratio'           : _modify_ratio,
    'timespan-eta'    : _modify_eta,
    'path'            : lambda raw: normpath(raw['downloadDir']),
    'trackers'        : TrackerList,
    'peers'           : PeerList,
    'files'           : _create_TorrentFileTree,
}

class Torrent(base.TorrentBase):
    """Information about a torrent as a mapping

    The available keys are specified in DEPENDENCIES and tkeys.TYPES.
    """

    def __init__(self, raw_torrent):
        self._raw = raw_torrent
        self._cache = {}

    def update(self, raw_torrent):
        old = self._raw
        cache = self._cache

        # Update an existing TorrentFileTree instead of creating a new one
        if 'fileStats' in raw_torrent and 'files' in cache:
            cache['files'].update(raw_torrent.pop('fileStats'))
            # Make sure the 'files' RPC field doesn't exist because it always
            # triggers the creation of a new TorrentFileTree (see
            # _create_TorrentFileTree).
            if 'files' in raw_torrent:
                del raw_torrent['files']

        # Remove cached values if their original/raw value(s) differ
        for k,v in tuple(cache.items()):
            fields = DEPENDENCIES[k]
            for field in fields:
                if field in old and field in raw_torrent and \
                   old[field] != raw_torrent[field]:
                    del cache[k]
                    break
        old.update(raw_torrent)

    def __getitem__(self, key):
        if key not in self._cache:
            if key in _MODIFY:
                # Modifier gets the whole raw torrent
                value = _MODIFY[key](self._raw)
            else:
                fields = DEPENDENCIES[key]
                assert len(fields) == 1
                try:
                    value = self._raw[fields[0]]
                except KeyError:
                    log.debug('%r not in %s', fields[0], tuple(self._raw))
                    raise KeyError(key)
            self._cache[key] = tkeys.TYPES[key](value)
        return self._cache[key]

    def __contains__(self, key):
        deps = DEPENDENCIES
        raw = self._raw
        # Check if key is known
        if key not in deps:
            return False
        else:
            # Check if we have all dependencies for key
            for dep in deps[key]:
                if dep not in raw:
                    return False
        return True

    def __iter__(self):
        for key in DEPENDENCIES:
            if key in self:
                yield key

    def __eq__(self, other):
        if isinstance(other, int):
            return self._raw['id'] == other
        elif isinstance(other, Torrent):
            return self._raw['id'] == other._raw['id']
        else:
            return NotImplemented

    def __lt__(self, other):
        return self._raw['id'] > other._raw['id']

    def __hash__(self):
        return hash(self._raw['id'])

    def clearcache(self):
        self._cache = {}


class TorrentFields(tuple):
    """Convert Torrent keys to those specified in rpc-spec.txt

    The resulting tuple has no duplicates and the keys 'id' and 'name' are
    always included.
    """
    _RPC_FIELDS = ('activityDate', 'addedDate', 'announceResponse', 'announceURL',
                   'bandwidthPriority', 'comment', 'corruptEver', 'creator',
                   'dateCreated', 'desiredAvailable', 'doneDate', 'downloadDir',
                   'downloadedEver', 'downloadLimit', 'downloadLimited',
                   'downloadLimitMode', 'error', 'errorString', 'eta', 'etaIdle',
                   'hashString', 'haveUnchecked', 'haveValid', 'honorsSessionLimits',
                   'id', 'isFinished', 'isPrivate', 'isStalled', 'lastAnnounceTime',
                   'lastScrapeTime', 'leftUntilDone', 'magnetLink',
                   'manualAnnounceTime', 'maxConnectedPeers',
                   'metadataPercentComplete', 'name', 'nextAnnounceTime',
                   'nextScrapeTime', 'peer-limit', 'peersConnected',
                   'peersGettingFromUs', 'peersSendingToUs', 'percentDone',
                   'pieceCount', 'pieceSize', 'queuePosition', 'rateDownload',
                   'rateUpload', 'recheckProgress', 'secondsDownloading',
                   'secondsSeeding', 'scrapeResponse', 'scrapeURL', 'seedIdleLimit',
                   'seedIdleMode', 'seedRatioLimit', 'seedRatioMode', 'sizeWhenDone',
                   'startDate', 'status', 'totalSize', 'torrentFile', 'uploadedEver',
                   'uploadLimit', 'uploadLimitMode', 'uploadLimited', 'uploadRatio',
                   'webseedsSendingToUs',

                   # Lists
                   'files', 'fileStats', 'peers', 'peersFrom', 'pieces', 'priorities',
                   'trackers', 'trackerStats', 'wanted', 'webseeds')

    _ALL_FIELDS = tuple(set(field
                            for fields in DEPENDENCIES.values()
                            for field in fields))
    _cache = {}

    def __new__(cls, *keys):
        if keys not in cls._cache:
            cls._cache[keys] = super().__new__(cls, cls._get_fields(*keys))
        return cls._cache[keys]

    @classmethod
    def _get_fields(cls, *keys):
        collected_fields = set(('id',))
        for key in keys:
            if key.lower() == 'all':
                return cls._ALL_FIELDS
            elif key in DEPENDENCIES:
                # key is one of Torrent's keys that needs one or more RPC fields
                collected_fields.update(DEPENDENCIES[key])
            elif key in cls._RPC_FIELDS:
                # key is a valid Transmission RPC field
                collected_fields.add(key)
            else:
                raise ValueError('Unknown torrent key: {!r}'.format(key))
        return collected_fields

    def __add__(self, other):
        if isinstance(other, (type(self), set, list, tuple)):
            fields = set(self)  # Make a copy
            fields.update(other)
            return type(self)(*fields)
        else:
            return NotImplemented

    def __eq__(self, other):
        return set(self) == set(other)

    def __ne__(self, other):
        return not self.__eq__(other)



# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details
# http://www.gnu.org/licenses/gpl-3.0.txt

"""Filtering TorrentFiles by various values"""

from ..tkeys import (TorrentFile, TorrentFilePriority)
from . import (BoolFilterSpec, CmpFilterSpec, Filter, FilterChain)


class SingleTorrentFileFilter(Filter):
    DEFAULT_FILTER = 'name'

    # Filters without arguments
    BOOLEAN_FILTERS = {
        'all': BoolFilterSpec(
            lambda f: True,
            aliases=('*',),
            description='All files'),
        'wanted': BoolFilterSpec(
            lambda f: f['is-wanted'],
            description='Wanted files'),
        'complete': BoolFilterSpec(
            lambda f: f['progress'] >= 100,
            description='Fully downloaded files'),
    }

    COMPARATIVE_FILTERS = {
        'name': CmpFilterSpec(
            lambda f, op, v: op(f['name'], v),
            description='Match VALUE against file name',
            value_type=TorrentFile.TYPES['name']),
        'size': CmpFilterSpec(
            lambda f, op, v: op(f['size-total'], v),
            aliases=(),
            description='Match VALUE against file size',
            value_type=TorrentFile.TYPES['size-total']),
        'downloaded': CmpFilterSpec(
            lambda f, op, v: op(f['size-downloaded'], v),
            aliases=('down',),
            description='Match VALUE against number of downloaded bytes',
            value_type=TorrentFile.TYPES['size-downloaded']),
        '%downloaded': CmpFilterSpec(
            lambda f, op, v: op(f['progress'], v),
            aliases=('%down',),
            description='Match VALUE against percentage of downloaded bytes',
            value_type=TorrentFile.TYPES['progress']),
        'priority': CmpFilterSpec(
            lambda f, op, v: op(f['priority'], v),
            aliases=('prio',),
            description='Match VALUE against download priority (low, normal, high or shun)',
            value_type=TorrentFile.TYPES['priority']),
    }


class TorrentFileFilter(FilterChain):
    """One or more filters combined with & and | operators"""
    filterclass = SingleTorrentFileFilter



# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details
# http://www.gnu.org/licenses/gpl-3.0.txt

import unicodedata
import re


def natsortkey(key):
    """Provide this as the 'key' argument to `list.sort`, `sorted`, etc.

    Pilfered from
    <https://blog.codinghorror.com/sorting-for-humans-natural-sort-order/>
    and adapted.
    """
    convert = lambda text: int(text) if isinstance(text, str) and text.isdigit() else text
    return [convert(c) for c in re.split('([0-9]+)', key)]


def striplines(lines):
    """Remove empty strings from start and end of `lines` using `pop`"""
    lines = list(lines)
    while lines and lines[0] == '':
        lines.pop(0)
    while lines and lines[-1] == '':
        lines.pop(-1)
    yield from lines


def strwidth(string):
    """Return displayed width of `string`, considering wide characters"""
    return len(string) + sum(1 for char in string
                             if unicodedata.east_asian_width(char) in 'FW')


def strcrop(string, width, tail=None):
    """Return `string` cropped to `width`, considering wide characters

    If `tail` is not None, it must be a string that is appended to the cropped
    string.
    """
    def widechar_indexes(s):
        for i,c in enumerate(s):
            if unicodedata.east_asian_width(c) in 'FW':
                yield i

    if strwidth(string) <= width:
        return string  # string is already short enough

    if tail is not None:
        width -= strwidth(tail)  # Account for tail in final width

    indexes = list(widechar_indexes(string)) + [len(string)]
    if not indexes:
        return string[:width]  # No wide chars, regular cropping is ok

    parts = []
    start = 0
    end = 0
    currwidth = strwidth(''.join(parts))

    while indexes and currwidth < width and end < len(string):
        end = indexes.pop(0)
        if end > 0:
            parts.append(string[start:end])
            currwidth = strwidth(''.join(parts))
            start = end

    if currwidth > width:
        excess = currwidth - width
        parts[-1] = parts[-1][:-excess]

    if tail is not None:
        parts.append(tail)

    return ''.join(parts)


def stralign(string, width, side='left'):
    """Return `string` aligned to `side`, considering wide characters

    The returned string is filled up with spaces to take `width` single-spaced
    characters and cropped with `strcrop` if necessary.
    """
    fill = width - strwidth(string)
    if fill < 0:
        string = strcrop(string, width)

    fill = width - strwidth(string)
    if fill > 0:
        if side == 'left':
            string = string + ' '*fill
        elif side == 'right':
            string = ' '*fill + string
        else:
            raise TypeError("side argument must be 'left' or 'right', not {!r}".format(side))
    return string



# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details
# http://www.gnu.org/licenses/gpl-3.0.txt

import sys
import asyncio
aioloop = asyncio.get_event_loop()


from . import cliopts
from . import logging
cliargs, clicmds = cliopts.parse()
if cliargs['profile_file'] is not None:
    logging.enable_profiling(cliargs['profile_file'])
logging.setup(debugmods=cliargs['debug'], filepath=cliargs['debug_file'])
logging.redirect_level('INFO', sys.stdout)
log = logging.make_logger()


from . import settings
cfg = settings.Settings()
settings.init_defaults(cfg)

def _log_cfg_change(setting):
    msg = '{} = {!s}'.format(setting.name, setting)
    if setting.value == setting.default:
        msg += ' (default)'
    log.debug(msg)
cfg.on_change(_log_cfg_change)


from .helpmgr import HelpManager
helpmgr = HelpManager()
helpmgr.settings = cfg


from .client import API
srvapi = API(url=cfg['srv.url'].value,
             interval=cfg['tui.poll'].value,
             loop=aioloop)
srvapi.bandwidth_unit = cfg['unit.bandwidth'].value
srvapi.bandwidth_prefix = cfg['unitprefix.bandwidth'].value
srvapi.size_unit = cfg['unit.size'].value
srvapi.size_prefix = cfg['unitprefix.size'].value

settings.init_server_defaults(cfg, srvapi.settings)


from .commands import CommandManager
cmdmgr = CommandManager(loop=aioloop)
cmdmgr.resources.update(aioloop=aioloop,
                        srvapi=srvapi,
                        cfg=cfg,
                        helpmgr=helpmgr)
helpmgr.commands = cmdmgr
cmdmgr.load_cmds_from_module(
    'stig.commands.cli', 'stig.commands.tui',
)


def run():
    from .commands.guess_ui import (guess_ui, UIGuessError)
    from . import hooks

    # Read commands from rc file
    rclines = ()
    if not cliargs['norcfile']:
        from .settings import rcfile
        from .settings.defaults import DEFAULT_RCFILE
        try:
            rclines = rcfile.read(cliargs['rcfile'] or DEFAULT_RCFILE)
        except rcfile.RcFileError as e:
            log.error('Loading rc file failed: {}'.format(e))
            sys.exit(1)

    # Decide if we run as a TUI or CLI
    if cliargs['tui']:
        cmdmgr.active_interface = 'tui'
    elif cliargs['notui']:
        cmdmgr.active_interface = 'cli'
    else:
        try:
            cmdmgr.active_interface = guess_ui(clicmds, cmdmgr)
        except UIGuessError as e:
            log.error('Unable to guess user interface')
            log.error('Provide one of these options: --tui/-t or --no-tui/-T')
            sys.exit(1)

    def run_commands():
        for cmdline in rclines:
            success = cmdmgr.run_sync(cmdline, on_error=log.error)
            # Ignored commands return None, which we consider a success here
            # because TUI commands like 'tab' in the rc file should have no
            # effect at all when in CLI mode.
            if success is False:
                return False

        # Exit if CLI commands fail
        if clicmds:
            success = cmdmgr.run_sync(clicmds, on_error=log.error)
            if not success:
                return False
        return True

    # Run commands either in CLI or TUI mode
    if cmdmgr.active_interface == 'cli':
        if not run_commands():
            sys.exit(1)
    elif cmdmgr.active_interface == 'tui':
        from .tui import main as tui
        cmdmgr.resources.update(tui=tui)
        if not tui.run(run_commands):
            sys.exit(1)

    # Terminate any remaining tasks
    tasks = tuple(task for task in asyncio.Task.all_tasks() if not task.done())
    if tasks:
        log.debug('Not all tasks have been properly canceled.')
        for task in tasks:
            log.debug('Terminating leftover task: %r', task)
            task.cancel()
            try:
                aioloop.run_until_complete(asyncio.wait_for(task, timeout=None))
                task.result()
            except (asyncio.CancelledError, asyncio.TimeoutError):
                pass

    aioloop.run_until_complete(srvapi.rpc.disconnect('Quit'))
    aioloop.close()



# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details
# http://www.gnu.org/licenses/gpl-3.0.txt

from functools import partial

from ..main import (cfg, srvapi, aioloop)
from . import main as tui
from ..client import ClientError

from .torrent.tlist import TorrentListWidget
from .torrent.flist import FileListWidget
from .torrent.plist import PeerListWidget

from .torrent.tlist_columns import TUICOLUMNS as TCOLUMNS
from .torrent.flist_columns import TUICOLUMNS as FCOLUMNS


def _connect_to_new_url(url):
    # See also ..hooks
    async def coro():
        try:
            await srvapi.rpc.connect(url.value)
        except ClientError as e:
            pass
    aioloop.create_task(coro())

cfg['srv.url'].on_change(_connect_to_new_url)


def _update_pollers(seconds):
    tui.srvapi.poll()

srvapi.rpc.on('connected', _update_pollers)


def _clear_list_widgets(seconds):
    from .torrent.tlist import TorrentListWidget
    from .torrent.flist import FileListWidget
    for widget in tui.tabs.contents:
        if isinstance(widget, (TorrentListWidget, FileListWidget)):
            widget.clear()

srvapi.rpc.on('disconnected', _clear_list_widgets)


def _set_poll_interval(seconds):
    tui.srvapi.interval = seconds.value

cfg['tui.poll'].on_change(_set_poll_interval)


def _set_cli_history(histfile):
    tui.cli.original_widget.history_file = histfile.value

cfg['tui.cli.history'].on_change(_set_cli_history)


def _set_autohide_delay(seconds):
    tui.log.autohide_delay = seconds.value

cfg['tui.log.autohide'].on_change(_set_autohide_delay)


def _set_log_maxrows(rows):
    tui.log.maxrows = rows.value

cfg['tui.log.height'].on_change(_set_log_maxrows)


def _set_theme(themefile):
    try:
        tui.load_theme(themefile.value)
    except tui.theme.ThemeError as e:
        raise ValueError(e)

cfg['tui.theme'].on_change(_set_theme)


def _refresh_tlists(value):
    for widget in tui.tabs:
        if isinstance(widget, (TorrentListWidget, FileListWidget, PeerListWidget)):
            widget.clear()
    tui.srvapi.poll()

cfg['unit.bandwidth'].on_change(_refresh_tlists)
cfg['unit.size'].on_change(_refresh_tlists)
cfg['unitprefix.bandwidth'].on_change(_refresh_tlists)
cfg['unitprefix.size'].on_change(_refresh_tlists)


def _set_tui_marked_char(methodname, setting):
    getattr(TCOLUMNS['marked'], methodname)(setting.value)
    getattr(FCOLUMNS['marked'], methodname)(setting.value)

    for widget in tui.tabs:
        if isinstance(widget, (TorrentListWidget, FileListWidget)):
            widget.refresh_marks()

cfg['tui.marked.on'].on_change(partial(_set_tui_marked_char, 'set_marked_char'), autoremove=False)
cfg['tui.marked.off'].on_change(partial(_set_tui_marked_char, 'set_unmarked_char'), autoremove=False)

_set_tui_marked_char('set_marked_char', cfg['tui.marked.on'])
_set_tui_marked_char('set_unmarked_char', cfg['tui.marked.off'])



# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details
# http://www.gnu.org/licenses/gpl-3.0.txt

"""Mixin classes for TUI commands"""

from ...logging import make_logger
log = make_logger(__name__)


from .. import ExpectedResource
from .. import utils


class polling_frenzy():
    aioloop = ExpectedResource

    def polling_frenzy(self, duration=2, short_interval=0.5):
        """Reduce polling interval to `short_interval` for `duration` seconds"""
        srvapi = self.srvapi
        if srvapi.interval > 1:
            import asyncio
            async def coro():
                log.debug('Setting poll interval to %s for %s seconds', short_interval, duration)
                orig_interval = srvapi.interval
                srvapi.interval = short_interval
                await asyncio.sleep(duration, loop=self.aioloop)
                srvapi.interval = orig_interval
                log.debug('Interval restored to %s', srvapi.interval)
            self.aioloop.create_task(coro())


class make_request():
    async def make_request(self, request_coro, polling_frenzy=False, quiet=False):
        """Awaits request coroutine and logs messages; returns response"""
        response = await request_coro
        utils.log_msgs(log, response.msgs, quiet)
        if response.success and polling_frenzy:
            self.polling_frenzy()
        return response


class select_torrents():
    tui = ExpectedResource

    def select_torrents(self, FILTER, allow_no_filter=True, discover_torrent=True):
        """Get TorrentFilter instance, focused torrent ID in a tuple or None

        If `FILTER` evaluates to True, it is passed to TorrentFilter and the
        resulting object is returned.

        If `FILTER` evaluates to False, a torrent ID is returned in a tuple if
        possible and `discover_torrent` evaluates to True. Otherwise, None is
        returned if `allow_no_filter` evaluates to True, else a ValueError is
        raised.

        Torrents are discovered by getting the:
            - `focused_torrent` attribute in this command's object (e.g. set
              by the 'tab' command),
            - `focused_torrent` attribute of the focused tab's widget
              (e.g. TorrentListWidget),
            - `focused_torrent_id` attribute of the focused tab's widget
              (e.g. FileListWidget).
        """
        if FILTER:
            from ...client import TorrentFilter
            return TorrentFilter(FILTER)
        else:
            if discover_torrent:
                tids = self._find_torrent_ids()
                if tids is not None:
                    return tids

            if allow_no_filter:
                return None

            else:
                raise ValueError('No torrent specified')

    def _find_torrent_ids(self):
        focused_widget = self.tui.tabs.focus

        # Get torrent ID from attribute set by 'tab' command (this happens if
        # for example when you run 'tab filelist' while a torrent is focused)
        if hasattr(self, 'focused_torrent'):
            return (self.focused_torrent['id'],)

        # Get torrent IDs from marks
        if hasattr(focused_widget, 'marked'):
            tids = tuple(twidget.torrent_id for twidget in focused_widget.marked)
            if tids:
                return tids

        # Get torrent ID from widget in focused tab (e.g. torrent lists)
        if hasattr(focused_widget, 'focused_torrent') and \
           focused_widget.focused_torrent is not None:
            return (focused_widget.focused_torrent.torrent_id,)

        # Get torrent ID from widget in focused tab (e.g. file lists)
        if hasattr(focused_widget, 'focused_torrent_id') and \
           focused_widget.focused_torrent_id is not None:
            return (focused_widget.focused_torrent_id,)


class select_files():
    tui = ExpectedResource

    def select_files(self, FILTER, allow_no_filter=True, discover_file=True):
        """Get TorrentFileFilter instance, focused file ID or None

        If `FILTER` evaluates to True, it is passed to TorrentFileFilter and
        the resulting object is returned.

        If `FILTER` evaluates to False and `discover_file` evaluates to True,
        the file ID (or IDs) are returned in a tuple if possible. Otherwise,
        None is returned if `allow_no_filter` evaluates to True, else a
        ValueError is raised.

        File are discovered by getting the `focused_file_id` of the focused
        tab's widget.
        """
        if FILTER:
            from ...client import TorrentFileFilter
            return TorrentFileFilter(FILTER)
        else:
            if discover_file:
                fids = self._find_file_ids()
                if fids is not None:
                    return fids
            if allow_no_filter:
                return None
            else:
                raise ValueError('No torrent file specified')

    def _find_file_ids(self):
        focused_widget = self.tui.tabs.focus

        if hasattr(focused_widget, 'marked'):
            tfids = tuple((fwidget.torrent_id, fwidget.file_id)
                          for fwidget in focused_widget.marked
                          if isinstance(fwidget.file_id, int))
            if tfids:
                return tfids

        # Get torrent ID from widget in focused tab
        if hasattr(focused_widget, 'focused_file_ids') and \
           focused_widget.focused_file_ids is not None:
            return tuple((focused_widget.focused_torrent_id, fid)
                         for fid in focused_widget.focused_file_ids)



# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details
# http://www.gnu.org/licenses/gpl-3.0.txt

"""Value types for Torrent classes"""

# The TYPES dictionary at the end of this file maps Torrent key names to
# types.  Every Torrent key must have a type, even if it's just a no-op
# (`lambda obj: obj`).
#
# A type is any callable that converts a single value to the appropriate class
# instance.
#
# Types are used to convert values from the server (e.g. large integers) and
# from the user (e.g. number strings like '3.5G').  Not all types must accept
# user-given values (e.g. 'files').


from ..logging import make_logger
log = make_logger(__name__)

from collections import abc
from .utils import pretty_float
import os


from itertools import chain
import re
class Number(float):
    """float with a nice string representation; also parses strings like '123K' or '123Mi'"""

    _PREFIXES_BINARY = (('Ti', 1024**4), ('Gi', 1024**3), ('Mi', 1024**2), ('Ki', 1024))
    _PREFIXES_METRIC = (('T', 1000**4), ('G', 1000**3), ('M', 1000**2), ('k', 1000))
    _ALL_PREFIXES = tuple((prefix.lower(), size)
                          for prefix,size in chain.from_iterable(zip(_PREFIXES_BINARY,
                                                                     _PREFIXES_METRIC)))
    _ALL_PREFIXES_DCT = dict(_ALL_PREFIXES)
    _REGEX = re.compile('^([-+]?\d+(?:\.\d+|)) ?(' +\
                        '|'.join(p[0] for p in _ALL_PREFIXES) + \
                        '|)(.*?)$',
                        flags=re.IGNORECASE)

    def __new__(cls, num, prefix='metric', unit=None):
        if isinstance(num, cls):
            return cls(float(num), prefix or num.prefix, unit or num.unit)

        elif isinstance(num, str):
            match = cls._REGEX.match(num)
            if match is None:
                raise ValueError('Not a number: {!r}'.format(num))
            else:
                num_str = match.group(1)
                unit = match.group(3) or unit
                prfx = match.group(2)
                num = float(num_str)
                if prfx:
                    all_prfxs = cls._ALL_PREFIXES_DCT
                    prfx_lower = prfx.lower()
                    if prfx_lower in all_prfxs:
                        num *= all_prfxs[prfx_lower]

                prfx_len = len(prfx)
                if prfx_len == 2:
                    prefix = 'binary'
                elif prfx_len == 1:
                    prefix = 'metric'

        obj = super().__new__(cls, num)
        if prefix == 'binary':
            obj._prefixes = cls._PREFIXES_BINARY
        elif prefix == 'metric':
            obj._prefixes = cls._PREFIXES_METRIC
        else:
            raise ValueError("prefix must be 'binary' or 'metric', not {!r}".format(prefix))
        obj.unit = unit
        obj.prefix = prefix
        return obj

    @property
    def with_unit(self):
        s = None
        for prefix,size in self._prefixes:
            if self >= size:
                s = '%s%s' % (pretty_float(self/size), prefix)
                break
        if s is None:
            s = pretty_float(self)
        if self.unit is not None:
            s += self.unit
        return s

    @property
    def without_unit(self):
        for prefix,size in self._prefixes:
            if self >= size:
                return pretty_float(self/size) + prefix
        return pretty_float(self)

    def __str__(self):
        return self.without_unit

    def __repr__(self):
        return '<{} {}, prefix={!r}, unit={!r}>'.format(type(self).__name__, float(self),
                                                        self.prefix, self.unit)

    def __add__(self, other): return type(self)(super().__add__(other))
    def __sub__(self, other): return type(self)(super().__sub__(other))
    def __mul__(self, other): return type(self)(super().__mul__(other))
    def __div__(self, other): return type(self)(super().__div__(other))
    def __truediv__(self, other): return type(self)(super().__truediv__(other))
    def __floordiv__(self, other): return type(self)(super().__floordiv__(other))
    def __mod__(self, other): return type(self)(super().__mod__(other))
    def __divmod__(self, other): return type(self)(super().__divmod__(other))
    def __pow__(self, other): return type(self)(super().__pow__(other))

# Because 'convert' needs Number, which is specified in this file, it must be
# imported AFTER Number exists to avoid a circular import.
from . import convert


class Percent(float):
    """Float with a pretty string representation"""
    def __str__(self):
        return pretty_float(self)


class Ratio(Number):
    """A Torrent's upload/download ratio as a float"""
    UNKNOWN = -1
    def __str__(self):
        if self == self.UNKNOWN:
            return '?'
        else:
            return pretty_float(self)


class SeedCount(Number):
    UNKNOWN = -1
    def __str__(self):
        return '?' if self == self.UNKNOWN else super().__str__()


class Status(str):
    """A Torrent's status as string"""
    VERIFY   = 'verifying'
    VERIFY_Q = 'verifying pending'
    LEECH    = 'leeching'
    LEECH_Q  = 'leeching pending'
    SEED     = 'seeding'
    SEED_Q   = 'seeding pending'
    STOPPED  = 'stopped'
    ORDER = (VERIFY, VERIFY_Q, LEECH, LEECH_Q, SEED, SEED_Q, STOPPED)

    def __new__(cls, status):
        if status not in cls.ORDER:
            raise ValueError('Invalid status string: {!r}'.format(status))
        else:
            obj = super().__new__(cls, status)
            obj._index = cls.ORDER.index(status)
            return obj

    def __lt__(self, other): return self._index < other._index
    def __le__(self, other): return self._index <= other._index
    def __gt__(self, other): return self._index > other._index
    def __ge__(self, other): return self._index >= other._index


import operator
import unicodedata
class SmartCmpStr(str):
    """String with smart comparison capabilities

    Adds the <, >, <=, >= operators that compare length of strings and makes
    comparison case-insensitive if the other string consists solely of
    lower-case characters.
    """

    def __new__(cls, string):
        # Combine characters with diacritical marks ("a˚" -> "å") so len()
        # reports the correct length.
        # http://www.unicode.org/faq/char_combmark.html
        return super().__new__(cls, unicodedata.normalize('NFC', string))

    def __cmp(self, other, op):
        if not isinstance(other, str):
            return NotImplemented

        # Do case-insensitive comparison?
        # Make copies to avoid infinite recursion.
        o = str(other)
        if o == o.lower():
            s = str(self.lower())
        else:
            s = str(self)

        if op in (operator.__eq__, operator.__ne__, operator.__contains__):
            return op(s, o)
        elif self.isdigit():
            return op(int(s), len(o))
        elif other.isdigit():
            return op(len(s), int(o))
        else:
            return op(s, o)

    def __lt__(self, other): return self.__cmp(other, operator.lt)
    def __le__(self, other): return self.__cmp(other, operator.le)
    def __eq__(self, other): return self.__cmp(other, operator.eq)
    def __ne__(self, other): return self.__cmp(other, operator.ne)
    def __gt__(self, other): return self.__cmp(other, operator.gt)
    def __ge__(self, other): return self.__cmp(other, operator.ge)
    def __contains__(self, other): return self.__cmp(other, operator.contains)

    def __hash__(self):
        return super().__hash__()



SECONDS = (('y', 31557600),  # 365.25 days
           ('M',  2629800),  # 1y / 12
           ('d',    86400),
           ('h',     3600),
           ('m',       60),
           ('s',        1))

class Timedelta(int):
    # To sort unknown and not applicable Timedeltas below the rest, these
    # constants have really large values that are very likely never going to
    # be of any use anyway.
    UNKNOWN        = 1e300
    NOT_APPLICABLE = 1e301
    assert UNKNOWN != NOT_APPLICABLE

    def __str__(self):
        if self == self.UNKNOWN:
            return '?'
        elif self == self.NOT_APPLICABLE:
            return ''
        elif -1 < self < 1:
            return 'now'

        abs_secs = abs(self)
        for i,(unit,amount) in enumerate(SECONDS):
            if abs_secs >= amount:
                num = self/amount

                # Small numbers get a sub-unit, for example '1d15h'
                if 1 <= abs_secs/amount < 10 and i < len(SECONDS)-1:
                    subunit, subamount = SECONDS[i+1]
                    if num >= 0:
                        subnum = abs( ((num%1) * amount) / subamount )
                    else:
                        subnum = abs( ((num%-1) * amount) / subamount )

                    if subnum >= 1:
                        return '%d%s%d%s' % (int(num), unit, int(subnum), subunit)

                return '%d%s' % (int(num), unit)

    @property
    def with_preposition(self):
        if self > 0:
            return 'in %s' % self
        elif self < 0:
            return ('%s ago' % self)[1:]  # Remove the first char ('-')
        else:
            return 'now'

    def __bool__(self):
        """Whether delta is known"""
        return self not in (self.UNKNOWN, self.NOT_APPLICABLE)


import time
class Timestamp(float):
    UNKNOWN = -1

    def __str__(self):
        if self == self.UNKNOWN:
            return 'sometime'
        abs_delta = abs(self - time.time())
        if abs_delta <= SECONDS[2][1]:      # 1 day: locale's time
            frmt = '%X'
        elif abs_delta <= SECONDS[2][1]*2:  # 2 days: locale's date and time
            frmt = '%x %X'
        else:                               # locale's date
            frmt = '%x'
        return time.strftime(frmt, time.localtime(self))

    def __bool__(self):
        """Whether timestamp is just a few seconds in the past/future"""
        return self != self.UNKNOWN

    @property
    def delta(self):
        if self == self.UNKNOWN:
            return Timedelta(Timedelta.UNKNOWN)
        else:
            return Timedelta(self - time.time())



from functools import total_ordering
@total_ordering
class TorrentFilePriority(str):
    INT2STR = {-1:'low', 0:'normal', 1:'high', -2:'shun'}
    STR2INT = {'low':-1, 'normal':0, 'high':1, 'shun':-2}

    def __new__(cls, prio):
        if isinstance(prio, int):
            if prio not in cls.INT2STR:
                raise ValueError('Invalid {} value: {!r}'.format(cls.__name__, prio))
            obj = super().__new__(cls, cls.INT2STR[prio])
        else:
            if prio not in cls.STR2INT:
                raise ValueError('Invalid {} value: {!r}'.format(cls.__name__, prio))
            obj = super().__new__(cls, prio)
        return obj

    def __int__(self): return self.STR2INT[self]
    def __lt__(self, other): return int(self) < int(other)
    def __repr__(self): return '<%s %r>' % (type(self).__name__, str(self))


import os
class TorrentFile(abc.Mapping):
    """Mapping that holds the values of a single file in a torrent"""

    # Distinguish subtrees from files without comparing classes everywhere
    nodetype = 'leaf'

    TYPES = {
        'tid'             : int,
        'id'              : int,
        'name'            : SmartCmpStr,
        'path'            : SmartCmpStr,
        'size-total'      : lambda val: convert.size(val, unit='byte'),
        'size-downloaded' : lambda val: convert.size(val, unit='byte'),
        'is-wanted'       : bool,
        'priority'        : TorrentFilePriority,
        'progress'        : Percent,
    }

    _VALUES = {
        'tid'             : lambda raw: raw['tid'],
        'id'              : lambda raw: raw['id'],
        'name'            : lambda raw: raw['name'],
        'path'            : lambda raw: os.sep.join(raw['path']),
        'size-total'      : lambda raw: raw['size-total'],
        'size-downloaded' : lambda raw: raw['size-downloaded'],
        'is-wanted'       : lambda raw: raw['is-wanted'],
        'priority'        : lambda raw: (TorrentFilePriority.STR2INT['shun']
                                         if not raw['is-wanted'] else raw['priority']),
        'progress'        : lambda raw: raw['size-downloaded'] / raw['size-total'] * 100,
    }

    def __init__(self, tid, id, name, path, size_total, size_downloaded, is_wanted, priority):
        self._raw = {'tid': tid, 'id': id, 'name': name, 'path': path,
                     'is-wanted': is_wanted, 'priority': priority,
                     'size-total': size_total, 'size-downloaded': size_downloaded}
        self._cache = {}

    def __getitem__(self, key):
        if key not in self._cache:
            self._cache[key] = self.TYPES[key](
                self._VALUES[key](self._raw)
            )
        return self._cache[key]

    def update(self, raw):
        self._raw.update(raw)
        try:
            for k in raw:
                del self._cache[k]
            if 'size-downloaded' in raw:
                del self._cache['progress']
        except KeyError:
            pass

    def __repr__(self): return '<{} {!r}>'.format(type(self).__name__, self['name'])
    def __iter__(self): return iter(self.TYPES)
    def __len__(self): return len(self.TYPES)


from . import base
def _ensure_TorrentFileTree(obj):
    if isinstance(obj, base.TorrentFileTreeBase):
        return obj
    else:
        raise RuntimeError('Not a TorrentFileTreeBase instance: {!r}'.format(obj))




import time
from collections import defaultdict
_PEER_PROGRESS_DATA = defaultdict(lambda: [])
MAX_SAMPLES = 5
MAX_SAMPLE_AGE = 3600
def gc_peer_progress_data():
    for peer_id,samples in tuple(_PEER_PROGRESS_DATA.items()):
        # Keep only the most recent samples
        while len(samples) > MAX_SAMPLES:
            samples.pop(0)

        # Also remove samples that are too old
        while samples and (samples[0][0] + MAX_SAMPLE_AGE) < time.monotonic():
            samples.pop(0)

        # Remove peer if there are no samples left
        if not samples:
            del _PEER_PROGRESS_DATA[peer_id]

def _guess_peer_rate_and_eta(peer_progress, peer_id, torrent_size):
    rate = 0
    if peer_progress >= 1:
        # Peer has already downloaded everything
        eta = Timedelta.NOT_APPLICABLE
    else:
        eta = Timedelta.UNKNOWN
        samples = _PEER_PROGRESS_DATA[peer_id]

        # Don't add the same progress twice
        if not samples or peer_progress != samples[-1][1]:
            samples.append((time.monotonic(), peer_progress))

        # We need at least 3 samples
        if len(samples) >= 3:
            # Use second and last sample to calculate rate.  The first sample is
            # ignored because its timestamp may be inaccurate: When we add the
            # first sample, the peer's progress is not current but the latest we
            # received, which happened likely tens of seconds ago.
            t_first, p_first = samples[1]
            t_last, p_last = samples[-1]
            p_diff = p_last - p_first
            t_diff = t_last - t_first

            # It's possible progress goes down, e.g. if a peer deletes a file
            if p_diff > 0:
                size_diff = torrent_size * p_diff
                rate = size_diff / t_diff
                size_remaining = torrent_size - (torrent_size * peer_progress)
                eta = size_remaining / rate

    return rate, eta

from . import geoip
class TorrentPeer(abc.Mapping):
    TYPES = {
        'id'        : lambda val: val,
        'tid'       : lambda val: val,
        'tname'     : SmartCmpStr,
        'tsize'     : lambda val: convert.size(val, unit='byte'),
        'ip'        : str,
        'port'      : int,
        'client'    : SmartCmpStr,
        'country'   : str,
        'progress'  : Percent,
        'rate-up'   : lambda val: convert.bandwidth(val, unit='byte'),
        'rate-down' : lambda val: convert.bandwidth(val, unit='byte'),
        'eta'       : Timedelta,
        'rate-est'  : lambda val: convert.bandwidth(val, unit='byte'),
    }

    _VALUES = {
        'id'      : lambda p: hash((p['tid'], p['ip'], p['port'])),
        'country' : lambda p: geoip.country_code(p['ip']) or '?',
    }

    def __init__(self, tid, tname, tsize, ip, port, client, progress, rate_up, rate_down):
        self._dct = {'tid': tid, 'tname': tname, 'tsize': tsize,
                     'ip': ip, 'port': port, 'client': client, 'progress': progress,
                     'rate-up': rate_up, 'rate-down': rate_down}
        self._cache = {}

    def __getitem__(self, key):
        if key not in self._cache:
            if key in ('eta', 'rate-est'):
                rate, eta = _guess_peer_rate_and_eta(
                    self['progress']/100, self['id'], self['tsize'])
                self._cache['rate-est'] = self.TYPES['rate-est'](rate)
                self._cache['eta'] = self.TYPES['eta'](eta)

            else:
                if key in self._VALUES:
                    val = self._VALUES[key](self._dct)
                else:
                    val = self._dct[key]
                self._cache[key] = self.TYPES[key](val)
        return self._cache[key]

    def __repr__(self): return '<{} {}@{}>'.format(type(self).__name__, self['ip'], self['progress'])
    def __iter__(self): return iter(self.TYPES)
    def __len__(self): return len(self.TYPES)



TYPES = {
    'id'                : int,
    'hash'              : str,
    'name'              : SmartCmpStr,
    'status'            : Status,
    'path'              : SmartCmpStr,
    'ratio'             : Ratio,

    'private'           : bool,
    'stalled'           : bool,
    'isolated'          : bool,

    '%downloaded'       : Percent,
    '%metadata'         : Percent,
    '%verified'         : Percent,

    'peers-connected'   : Number,
    'peers-uploading'   : Number,
    'peers-downloading' : Number,
    'peers-seeding'     : SeedCount,

    'timestamp-created' : Timestamp,
    'timestamp-added'   : Timestamp,
    'timestamp-started' : Timestamp,
    'timestamp-active'  : Timestamp,
    'timestamp-done'    : Timestamp,
    'timespan-eta'      : Timedelta,
    'timestamp-manual-announce-allowed': lambda v: Timestamp(v) if v > 0 else Timestamp(Timestamp.UNKNOWN),

    'rate-down'         : lambda v: convert.bandwidth(v, unit='byte'),
    'rate-up'           : lambda v: convert.bandwidth(v, unit='byte'),

    'size-final'        : lambda v: convert.size(v, unit='byte'),
    'size-total'        : lambda v: convert.size(v, unit='byte'),
    'size-downloaded'   : lambda v: convert.size(v, unit='byte'),
    'size-uploaded'     : lambda v: convert.size(v, unit='byte'),
    'size-available'    : lambda v: convert.size(v, unit='byte'),
    'size-corrupt'      : lambda v: convert.size(v, unit='byte'),

    'trackers'          : tuple,
    'files'             : _ensure_TorrentFileTree,
    'peers'             : tuple,
}



# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details
# http://www.gnu.org/licenses/gpl-3.0.txt

"""Column specifications for tables"""

from ..logging import make_logger
log = make_logger(__name__)

from ..utils import stralign


class ColumnBase():
    header = {'left': '', 'right': ''}
    width = None
    align = 'right'
    interfaces = ('cli', 'tui')
    may_have_wide_chars = False

    def __init__(self, data=None):
        self.data = data if data is not None else {}
        super().__init__()

    def get_value(self):
        raise NotImplementedError()

    def get_raw(self):
        return self.get_value()

    def get_string(self):
        """Return `get_value` as spaced and aligned string

        If the `width` attribute is not set to None, expand or shrink and
        align the returned string (`align` attribute must be 'left' or
        'right').
        """
        width = self.width
        align = self.align
        string = str(self.get_value())
        if not isinstance(width, int):
            return string
        else:
            # Crop string or fill it with spaces and align to left/right
            if self.may_have_wide_chars:
                return stralign(string, width, align)

            if len(string) > width:
                string = string[:width]

            if align == 'right':
                return string.rjust(width)
            elif align == 'left':
                return string.ljust(width)
            else:
                raise TypeError("'align' attribute must be 'left' or 'right', not {!r}".format(align))

    def __repr__(self):
        if self.data:
            return '<{} {}>'.format(type(self).__name__, self.get_value())
        else:
            return '<{} <UNINITIALIZED>>'.format(type(self).__name__)



# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details
# http://www.gnu.org/licenses/gpl-3.0.txt

"""Converters for units"""

from ..logging import make_logger
log = make_logger(__name__)

from .tkeys import Number


class _DataCountConverter():
    """Convert bits or bytes with metric or binary unit prefix to Number"""

    def __init__(self):
        self.prefix = 'metric'
        self.unit = 'byte'

    # Units (Long2Short and Short2Long)
    UNITS_L2S = {'bit': 'b', 'byte': 'B', 'b': 'b', 'B': 'B'}
    UNITS_S2L = {'b': 'bit', 'B': 'byte', 'bit': 'bit', 'byte': 'byte'}

    def __call__(self, num, unit=None):
        """Make Number from `num`

        `num` can be a number (int, float) or a string.  If it's a string, it
        can specify the unit ('b' for bits, 'B' for bytes), in which case the
        `unit` argument is ignored.

        If `unit` is not specified, it defaults to whatever the `unit`
        property is currently set to.
        """
        n = Number(num, prefix=self._prefix, unit=self.UNITS_L2S.get(unit, None))
        unit = n.unit
        if unit is None:
            unit = self.unit
            n.unit = self.UNITS_L2S[unit]
        elif unit not in self.UNITS_S2L:
            raise ValueError("Unit must be 'b' (bit) or 'B' (byte), not {!r}".format(unit))

        # Maybe we need to convert from bits to bytes or vice versa
        if self.UNITS_S2L[unit] == self._unit:
            return n
        else:
            return Number(self._convert(n), prefix=self._prefix,
                          unit=self.UNITS_L2S[self._unit])

    @property
    def unit(self):
        """'bit' or 'byte'"""
        return self._unit

    @unit.setter
    def unit(self, unit):
        if unit == 'bit':
            # How to get bits if we got bytes
            self._convert = lambda num: num*8
        elif unit == 'byte':
            # How to get bytes if we got bits
            self._convert = lambda num: num/8
        else:
            raise ValueError("unit must be 'bit' or 'byte'")
        self._unit = unit

    @property
    def prefix(self):
        """'binary' or 'metric'"""
        return self._prefix

    @prefix.setter
    def prefix(self, prefix):
        if prefix not in ('binary', 'metric'):
            raise ValueError("unit must be 'binary' or 'metric'")
        else:
            self._prefix = prefix


bandwidth = _DataCountConverter()

class bandwidth_mixin():
    @property
    def bandwidth_unit(self):
        """Bandwidth numbers are displayed in 'bit' or 'byte'"""
        return bandwidth.unit

    @bandwidth_unit.setter
    def bandwidth_unit(self, unit):
        bandwidth.unit = unit

    @property
    def bandwidth_prefix(self):
        """Bandwidth numbers are displayed with 'binary' or 'metric' unit prefixes"""
        return bandwidth.prefix

    @bandwidth_prefix.setter
    def bandwidth_prefix(self, prefix):
        bandwidth.prefix = prefix


size = _DataCountConverter()

class size_mixin():
    @property
    def size_unit(self):
        """Size numbers are displayed in 'bit' or 'byte'"""
        return size.unit

    @size_unit.setter
    def size_unit(self, unit):
        size.unit = unit

    @property
    def size_prefix(self):
        """Size numbers are displayed with 'binary' or 'metric' unit prefixes"""
        return size.prefix

    @size_prefix.setter
    def size_prefix(self, prefix):
        size.prefix = prefix



# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details
# http://www.gnu.org/licenses/gpl-3.0.txt

from ..logging import make_logger
log = make_logger(__name__)

import re


class Key(str):
    """Key or key combination as string"""

    # Convert some urwid key names and some other stuff
    _INIT = (
        (re.compile(r'^<(.+)>$'),                    r'\1'),
        (re.compile(r'^(.*) $'),                     r'\1space'),
        (re.compile(r'^meta', flags=re.I),           r'alt'),
        (re.compile(r'(^|-)esc$', flags=re.I),       r'\1escape'),
        (re.compile(r'(^|-)pos1$', flags=re.I),      r'\1home'),
        (re.compile(r'(^|-)del$', flags=re.I),       r'\1delete'),
        (re.compile(r'(^|-)ins$', flags=re.I),       r'\1insert'),
        (re.compile(r'(^|-)return$', flags=re.I),    r'\1enter'),
        (re.compile(r'(^|-)page up$', flags=re.I),   r'\1pgup'),
        (re.compile(r'(^|-)page down$', flags=re.I), r'\1pgdn'),
        (re.compile(r'(^|-)page dn$', flags=re.I),   r'\1pgdn'),
        (re.compile(r' '),                           r'-'),
        # The first part in key combos must always be the same, but the part
        # after must be preserved. <alt-l> is not the same as <alt-L>.
        (re.compile(r'^(\w+)-(\S+)$'),
         lambda m: m.group(1).lower()+'-'+m.group(2)),
    )

    _MODS = ('shift', 'alt', 'ctrl')
    _FKEYS = ('F1', 'F2', 'F3', 'F4', 'F5', 'F6', 'F7', 'F8', 'F9', 'F10', 'F11', 'F12',
              'F13', 'F14', 'F15', 'F16', 'F17', 'F18', 'F19', 'F20')
    _KEYNAMES = ('escape', 'space', 'home', 'end', 'tab', 'delete', 'backspace', 'insert',
                 'enter', 'pgup', 'pgdn', 'left', 'right', 'up', 'down') + _FKEYS
    _cache = {}

    def __new__(cls, key):
        if isinstance(key, Key):
            return key
        elif key in cls._cache:
            return cls._cache[key]
        else:
            orig_key = key

        # Remove brackets (<>) around key, some renaming, etc.
        for regex,repl in cls._INIT:
            key = regex.sub(repl, key)

        # Split modifiers and key
        if '-' in key and len(key) > 1:
            _parts = key.split('-')
            key = _parts.pop(-1)
            mods = set(p.lower() for p in _parts)
            for mod in mods:
                if mod not in cls._MODS:
                    raise ValueError('Invalid modifier in <%s>: <%s>' % (orig_key, mod))

            if len(key) == 0:
                raise ValueError('Missing key after modifier: <%s>' % orig_key)

            # Convert <shift-x> to <X>
            elif len(key) == 1 and 'shift' in mods:
                key = key.upper()
                mods.remove('shift')

            # 'shift-/ctrl-E' is the same as 'shift-/ctrl-e'
            elif 'alt' not in mods:
                key = key.lower()
        else:
            mods = set()

        # Validate named keys ('enter', 'delete', etc)
        if len(key) > 1:
            # F* key names are upper case, all others lower case
            key_upper = key.upper()
            key = key_upper if key_upper in cls._FKEYS else key.lower()
            if key not in cls._KEYNAMES:
                raise ValueError('Unknown key name: %r' % key)

        if mods:
            keystr = '%s-%s' % ('-'.join(sorted(mods)), key)
        else:
            keystr = key

        if len(key) < 1:
            raise ValueError('No key specified')

        obj = super().__new__(cls, keystr)
        cls._cache[orig_key] = obj
        return obj

    def __str__(self):
        return '<%s>' % super().__str__()

    def __repr__(self):
        return '<Key %s>' % self


class KeyChain(tuple):
    class COMPLETED(): pass
    class ADVANCED(): pass
    class ABORTED(): pass
    class REFUSED(): pass

    def __new__(cls, *keys):
        obj = super().__new__(cls, (Key(k) for k in keys))
        if len(keys) < 2:
            raise ValueError('Not enough keys to make a chain: %s' % str(obj))
        return obj

    def __init__(self, *keys):
        self._pos = -1

    def feed(self, key):
        """Process `key`

        If `key` is expected, `advance` is called.  If that completes the
        chain, `reset` is called.

        If `key` is not expected, `reset` called.

        Return one of the constants:
          - COMPLETED: chain was completed with `key`
          - ADVANCED: chain was advanced with `key` but is not complete yet
          - REFUSED: `key` != `next_key` and `given` is empty
          - ABORTED: `key` != `next_key` and `given` is not empty
        """
        if self.next_key == key:
            log.debug('Advancing %r with %r', self, key)
            self.advance()
            if self.is_complete:
                self.reset()
                return self.COMPLETED
            else:
                return self.ADVANCED
        elif self._pos == -1:
            return self.REFUSED
        else:
            # log.debug('Expected %r, got %r', self.next_key, key)
            self.reset()
            return self.ABORTED

    def advance(self):
        """Look for next key in chain or call `reset` if complete"""
        if self.is_complete:
            log.debug('Completed %r', self)
            self.reset()
        else:
            self._pos += 1

    def reset(self):
        """Start looking for first key in chain"""
        self._pos = -1

    @property
    def next_key(self):
        """Next missing key or None if chain is already complete"""
        return None if self.is_complete else self[self._pos+1]

    def startswith(self, keys):
        """Whether this key chain starts with `keys` (sequence of Key instances)"""
        return self[:len(keys)] == tuple(keys)

    @property
    def given(self):
        """Tuple of keys we have so far"""
        return tuple(self) if self.is_complete else self[:self._pos+1]

    @property
    def is_complete(self):
        """Wether there are keys missing to complete the chain"""
        return self._pos >= len(self)-1

    @property
    def is_started(self):
        """Wether the first character has been supplied"""
        return self._pos > -1

    def __str__(self):
        return ' '.join(str(k) for k in self)

    def __repr__(self):
        text = ['<KeyChain %s' % str(self)]
        if self._pos > -1:
            text.append(' given=[')
            text.append(' '.join(str(k) for k in self.given))
            text.append(']')
        text.append('>')
        return ''.join(text)



class KeyMapped():
    """Mixin class that provides a keypress method"""
    def keypress(self, size, key):
        key = Key(key)

        sup = super()
        def try_super(key):
            if sup.selectable():
                key = sup.keypress(size, key)
            return key

        # Offer key to parent's keypress() for built-in keys like up/down
        key = try_super(key)

        # If parent doesn't want key, check if it is mapped
        if key is not None:
            key_eval = self.keymap.evaluate(key,
                                            context=self.context,
                                            callback=self.callback,
                                            widget=self)
            if key_eval is None:
                return None
            elif key_eval != key:
                # evaluate() may have resolved key to another key
                # (e.g. 'j' -> 'down')
                key = try_super(key_eval)

        return key

    def selectable(self):
        return True


NOCONTEXT = object()  # Another None value

class KeyMap():
    """Bind keys to actions in different contexts

    To bind keys to widgets, a new class is created that accepts user input by
    adding the methods `keypress` and `selectable` to the original class:

    >>> keymap = KeyMap()
    >>> keymap.bind(key='meta g',
                    action=lambda widget: widget.set_edit_text('abcfoo'),
                    context='pwentry')
    >>> Password = keymap.wrap(urwid.Edit, context='pwentry')
    >>> Password
    urwid.Edit_KeyMapped

    Now you can use Password like any regular Edit widget:

    >>> pw = Password('', 'Enter password or <Alt-g> to generate one')
    """

    def __init__(self, callback=None):
        self._callback = callback
        self._contexts = {None: {}}

        self._keychain_callbacks = []
        self._keychain_partial = []
        self._keychain_context = NOCONTEXT

    def bind(self, key, action, context=None):
        """Bind `key` to `action` in `context`

        key: a key string that is converted to Key or KeyChain with `mkkey`
        context: a descriptive string (or any other hashable object)
        action: What to do when `key` is pressed in `context`; may be:
                - a callable that accepts the widget that received `key`,
                - a Key object (so you can bind <j> to <down> for example),
                - any other object that is passed to the `callback` specified
                  in calls to `wrap` or when the KeyMap was instantiated.
        """
        key = self.mkkey(key)
        if isinstance(action, Key) and key == action:
            raise ValueError('Mapping {} to {} is silly'.format(key, action))

        if context not in self._contexts:
            self._contexts[context] = {}
        self._contexts[context][key] = action
        log.debug('%s: Mapped %r -> %r', context, key, action)

    def unbind(self, key, context=None):
        """Unbind `key` in `context`

        Key chains starting with `key` are also removed.
        """
        key = self.mkkey(key)
        if context not in self._contexts:
            raise ValueError('Unknown context: {!r}'.format(context))
        elif key in self._contexts[context]:
            del self._contexts[context][key]
            log.debug('Unmapped %r [%s]', key, context)
        else:
            key_removed = False
            for k in tuple(self._contexts[context]):
                if isinstance(k, KeyChain) and k.startswith(key):
                    del self._contexts[context][k]
                    log.debug('Unmapped %r [%s]', k, context)
                    key_removed = True
            if not key_removed:
                raise ValueError('Key not mapped in context {!r}: {}'.format(context, key))

    def evaluate(self, key, context=None, callback=None, widget=None):
        """Run action that is bound to `key` in `context`

        key: the pressed key
        context: the context `key` was pressed in
        callback: None or any callable
        widget: the widget that received `key`
        """
        if context not in self._contexts:
            raise ValueError('Unknown context: {}'.format(context))

        key = Key(key)
        action = None
        log.debug('Evaluating %r in context %r for widget %r', key, context, widget)

        # Unless we've are already started a key chain, try to find a single-key mapping
        if not self._keychain_partial:
            action = self._get_single_key_action(key, context)
        else:
            log.debug('Not doing single-key lookup because we\'re '
                      'trying to complete a key chain')

        # Try to advance keychains only if no keychain was started previously
        # or if that previously started keychain was in the same context as
        # we're in now.
        if action is None and (self._keychain_context == NOCONTEXT or
                               self._keychain_context == context):
            action = self._get_keychain_action(key, context, self._keychain_partial)
            if action is KeyChain.ADVANCED:
                self._keychain_context = context    # Lock context
                self._keychain_partial.append(key)
                log.debug('%r was used to advance a keychain (status: %r)', key, self._keychain_partial)
                self._run_callbacks(tuple(self._active_keychains(context, self._keychain_partial)))
                return None
            elif action is KeyChain.ABORTED:
                log.debug('%r was used to abort a keychain', key)
                self._reset_keychains(context, self._keychain_partial)
                self._keychain_context = NOCONTEXT
                self._keychain_partial.clear()
                self._run_callbacks(tuple())
                return None
            elif action is KeyChain.REFUSED:
                log.debug('%r was refused by all keychains', key)
                self._keychain_context = NOCONTEXT
                self._keychain_partial.clear()
                action = None
            elif isinstance(action, Key):
                log.debug('%r was resolved to a single key: %r', key)
                self._keychain_context = NOCONTEXT
                self._keychain_partial.clear()
                self._run_callbacks(tuple())
            else:
                log.debug('%r was used to complete a keychain', key)
                self._keychain_context = NOCONTEXT
                self._keychain_partial.clear()
                self._run_callbacks(tuple())

        # Handle the action we found
        log.debug('Evaluated %r to action: %r', key, action)
        if action is None:
            log.debug('%r is not mapped in context %r', key, context)
            return key
        elif isinstance(action, Key):
            log.debug('Evaluating %r', action)
            return self.evaluate(action, context, callback, widget)
        elif isinstance(action, KeyChain):
            log.error('Mapping to key chains is not supported.')
            return None
        elif callable(action):
            # action itself is the callback
            log.debug('Calling action: %r', action)
            action(widget)
        elif callback is not None:
            # Individual callback for this widget
            log.debug('Calling widget callback %r with widget %r', callback, widget)
            callback(action, widget)
        elif self._callback is not None:
            # General callback for all widgets
            log.debug('Calling default callback %r with widget %r', self._callback, widget)
            self._callback(action, widget)
        else:
            raise TypeError('No callback given - unable to handle {!r}'.format(action))

    def _get_single_key_action(self, key, context):
        for keymap in (self._contexts[context], self._contexts[None]):
            if key in keymap:
                return keymap[key]

    def _get_keychain_action(self, key, context, partial_keys):
        # Go through all actions in this context and feed them the new key.
        action = KeyChain.ABORTED if partial_keys else KeyChain.REFUSED
        for kc,act in self._keychains(context, partial_keys):
            log.debug('   Feeding %r to %r', key, kc)
            result = kc.feed(key)

            # The first completed chain wins, we return its action and reset
            # all other chains.
            if result is KeyChain.COMPLETED:
                log.debug('%r completed %r', key, kc)
                self._reset_keychains(context, partial_keys)
                return act

            # At least this key chain was advanced.
            elif result is KeyChain.ADVANCED:
                log.debug('%r advanced %r', key, kc)
                action = KeyChain.ADVANCED

        if action is KeyChain.ADVANCED:
            log.debug('At least one key chain was advanced, returning %r', action)
        elif action is KeyChain.ABORTED:
            log.debug('No key chain was advanced or completed, returning %r', action)
        elif action is KeyChain.REFUSED:
            log.debug('%r did not start or advance a keychain, returning %r', key, action)
        else:
            log.debug('There should be no keychains in context %r: %r', context, tuple(self._keychains(context)))

        return action

    def _reset_keychains(self, context, partial_keys):
        for kc,action in self._keychains(context, partial_keys):
            kc.reset()

    def _keychains(self, context, partial_keys):
        for kc,action in self._contexts[context].items():
            if isinstance(kc, KeyChain) and kc.startswith(partial_keys):
                yield (kc, action)

    def _active_keychains(self, context, partial_keys):
        for kc,action in self._keychains(context, partial_keys):
            if kc.is_started:
                yield (kc, action)

    def _run_callbacks(self, active_keychains):
        log.debug('Running callbacks with %r', active_keychains)
        for cb in self._keychain_callbacks:
            cb(active_keychains)

    def on_keychain(self, callback):
        self._keychain_callbacks.append(callback)

    def wrap(self, cls, callback=None, context=None):
        """Return widget class that passes keypress()es through `evaluate`

        callback: Custom callable for this widget class; will be called with
                  the action bound to the key and the widget instance.
        context: A descriptive string (or any other hashable object)
        """
        if context not in self._contexts:
            self._contexts[context] = {}

        log.debug('Wrapping %r in %r with context %r', cls, KeyMapped, context)
        clsname = cls.__name__ + '_KeyMapped'
        return type(clsname,
                    (KeyMapped, cls),
                    {'keymap': self,
                     'context': context,
                     'callback': staticmethod(callback)})

    _KEY_SPLIT_SPACE = re.compile(r' +')
    def mkkey(self, string):
        """Create Key or KeyChain instance from `string`

        TODO: describe format
        """
        if isinstance(string, (Key, KeyChain)):
            return string
        elif ' ' in string:
            keys = (key for key in self._KEY_SPLIT_SPACE.split(string) if key)
            return KeyChain(*keys)
        elif '+' in string and len(string) >= 3:
            return KeyChain(*string.strip().split('+'))
        else:
            return Key(string)

    def keys(self, match_func=None, context=None):
        """Yield all keys where `match_func(key, action)` evaluates to True

        If `context` is not None, yield only keys mapped in that context.
        """
        def is_match(key, action):
            return match_func is None or match_func(key, action)

        if context is not None:
            for k,a in self._contexts[context].items():
                if is_match(k, a):
                    yield k
        else:
            for context in self.contexts:
                for k,a in self._contexts[context].items():
                    if is_match(k, a):
                        yield k

    def map(self, context=None):
        """Yield (key, action) tuples for keys mapped in `context`"""
        yield from self._contexts[context].items()

    @property
    def contexts(self):
        """Yield existing contexts"""
        yield from self._contexts


from urwid.command_map import (command_map, REDRAW_SCREEN, CURSOR_UP,
                               CURSOR_DOWN, CURSOR_LEFT, CURSOR_RIGHT,
                               CURSOR_PAGE_UP, CURSOR_PAGE_DOWN,
                               CURSOR_MAX_LEFT, CURSOR_MAX_RIGHT, ACTIVATE)
CANCEL = 'cancel'
CURSOR_WORD_LEFT = 'cursor word left'
CURSOR_WORD_RIGHT = 'cursor word right'
DELETE_TO_EOL = 'delete to end of line'
DELETE_LINE = 'delete line'
DELETE_CHAR_UNDER_CURSOR = 'delete char under cursor'
DELETE_WORD_LEFT = 'delete word left'
DELETE_WORD_RIGHT = 'delete word right'

# Remove urwid's defaults
for key in tuple(command_map._command):
    del command_map[key]

command_map[Key('pgup')]   = CURSOR_PAGE_UP
command_map[Key('pgdn')]   = CURSOR_PAGE_DOWN
command_map[Key('ctrl-b')] = CURSOR_PAGE_UP
command_map[Key('ctrl-f')] = CURSOR_PAGE_DOWN

command_map[Key('up')]     = CURSOR_UP
command_map[Key('down')]   = CURSOR_DOWN
command_map[Key('left')]   = CURSOR_LEFT
command_map[Key('right')]  = CURSOR_RIGHT
command_map[Key('meta-b')] = CURSOR_WORD_LEFT
command_map[Key('meta-f')] = CURSOR_WORD_RIGHT

command_map[Key('home')]   = CURSOR_MAX_LEFT
command_map[Key('end')]    = CURSOR_MAX_RIGHT
command_map[Key('ctrl-a')] = CURSOR_MAX_LEFT
command_map[Key('ctrl-e')] = CURSOR_MAX_RIGHT

command_map[Key('ctrl-k')] = DELETE_TO_EOL
command_map[Key('ctrl-u')] = DELETE_LINE
command_map[Key('ctrl-d')] = DELETE_CHAR_UNDER_CURSOR
command_map[Key('meta-d')] = DELETE_WORD_LEFT
command_map[Key('meta-backspace')] = DELETE_WORD_RIGHT
command_map[Key('ctrl-w')] = DELETE_WORD_RIGHT

command_map[Key('enter')]  = ACTIVATE
command_map[Key('escape')] = CANCEL
command_map[Key('ctrl-g')] = CANCEL
command_map[Key('ctrl-l')] = REDRAW_SCREEN



# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details
# http://www.gnu.org/licenses/gpl-3.0.txt

from ..logging import make_logger
log = make_logger(__name__)

import asyncio
import functools
import blinker

from . import errors
from .utils import (SleepUneasy, PerfectInterval)


def _funcname(func, *posargs, **kwargs):
    if hasattr(func, '__qualname__'):
        name = func.__qualname__
    elif hasattr(func, '__name__'):
        name = func.__name__
    else:
        name = repr(func)

    args = []
    if posargs:
        args.append(', '.join(repr(arg) for arg in posargs))
    if kwargs:
        args.append(', '.join('%s=%r' % (k,v) for k,v in kwargs.items()))

    name += '(' + ', '.join(args) + ')'
    return name


class RequestPoller():
    """Continuously send request and publish the response

    request: Coroutine that is called at intervals
    interval: Delay between calls
    loop: Asyncio loop

    Any other positional or keyword arguments are passed to `request`.
    """
    def __init__(self, request, *args, interval=1, loop=None, **kwargs):
        self.loop = loop if loop is not None else asyncio.get_event_loop()
        self._on_response = blinker.Signal()
        self._on_error = blinker.Signal()
        self._prev_error = None
        self._interval = interval
        self._poll_task = None
        self._poll_loop_task = None
        self._sleep = SleepUneasy(loop=loop)
        self._skip_ongoing_request = False
        self.set_request(request, *args, **kwargs)

    async def start(self):
        """Start polling"""
        if self.running:
            log.debug('Already polling: %s', self._request_str)
        else:
            log.debug('Starting polling: %s', self._request_str)
            self._poll_loop_task = self.loop.create_task(self._poll_loop())
            def reraise(task):
                log.debug('Polling loop is finished: %s', self._request_str)
                # Ignore if _poll_loop() was cancelled, raise all other exceptions
                try:
                    task.result()
                except asyncio.CancelledError:
                    pass
            self._poll_loop_task.add_done_callback(reraise)

    async def _poll_loop(self):
        self._prev_error = None
        while True:
            pinterval = PerfectInterval(self.loop)

            self._poll_task = self.loop.create_task(self._do_poll())
            try:
                await self._poll_task
            except asyncio.CancelledError:
                if self._skip_ongoing_request:
                    log.debug('Skipping polling result once: %s', self._request_str)
                    continue  # Skip the sleep() at the end of the loop
                else:
                    raise
            else:
                log.debug('Finished polling gracefully: %s', self._request_str)
            finally:
                self._poll_task = None
                self._skip_ongoing_request = False

            await self._sleep.sleep(pinterval(self._interval))

    async def _do_poll(self):
        """Send request and send response or error

        The return value from the request is passed to the 'response' event
        handlers.

        Exceptions raised by the request are passed to the 'error' handlers,
        and, except for `ConnectionError`, `stop` is called.
        """
        if self._request is None:
            log.debug('Request is set to None - not requesting anything')
        else:
            log.debug('Polling: %s', self._request_str)
            try:
                response = await self._request()
            except asyncio.CancelledError:
                # In _poll_loop(), abort waiting for poll_task.
                raise
            except errors.ClientError as e:
                # Report error but keep trying to connect
                self._run_callbacks(error=e)
            else:
                self._run_callbacks(response=response)

    def _run_callbacks(self, response=None, error=None):
        if self._skip_ongoing_request:
            log.debug('Not running callbacks for: %s', self._request_str)
            self._skip_ongoing_request = False
        else:
            log.debug('Running callbacks for: %s', self._request_str)
            self._on_response.send(response)
            # Ignore duplicate errors
            if error is not None and str(self._prev_error) != str(error):
                self._prev_error = error
                if self._on_error.receivers:
                    self._on_error.send(error)
                else:
                    log.debug('Uncaught exception in %r', self)
                    raise error

    def skip_ongoing_request(self):
        """Stop a currently ongoing request; do nothing if there is no ongoing request"""
        if self._poll_task is not None:
            log.debug('Skipping ongoing request: %s', self._request_str)
            self._skip_ongoing_request = True
            self._poll_task.cancel()

    async def stop(self):
        """Stop polling"""
        if not self.running:
            log.debug('Already stopped polling: %s', self._request_str)
        else:
            log.debug('Stopping polling %s', self._request_str)
            self._poll_loop_task.cancel()
            try:
                await asyncio.wait_for(self._poll_loop_task, timeout=0, loop=self.loop)
            except (asyncio.CancelledError, asyncio.TimeoutError):
                pass
            finally:
                self._poll_loop_task = None
                self._run_callbacks()

    def __del__(self):
        if self._poll_loop_task is not None:
            raise RuntimeError('You forgot to call stop() on {!r}'.format(self))

    def poll(self):
        """Poll immediately instead of waiting for next interval

        This also resets the interval - the next request is made `interval`
        seconds after this method is called.

        Do nothing if this poller is not started.
        """
        if self.running:
            self._sleep.interrupt()

    @property
    def running(self):
        """Whether poller is polling"""
        return self._poll_loop_task is not None

    def set_request(self, request, *args, **kwargs):
        """Set request to coroutine `request`

        Any positional or keyword arguments are passed to `request` at each
        call.
        """
        self._request_str = _funcname(request, *args, **kwargs)
        log.debug('Setting new request: %s', self._request_str)
        if args or kwargs:
            self._request = functools.partial(request, *args, **kwargs)
        else:
            self._request = request

    @property
    def request(self):
        """The request coroutine sent every interval

        If any arguments were given, this is a partial object.
        """
        return self._request

    def on_response(self, callback, autoremove=True):
        """Register `callback` to receive responses

        callback: Any callable that gets the return value from the request
        autoremove: Store callback as weak reference and remove it
                    automatically once there are no other references to it
                    left

        If the request raises an exception, 'response' callbacks are called
        with `None` and 'error' callbacks are called with the exception.
        """
        log.debug('Registering %r to receive %s responses', _funcname(callback), self._request_str)
        self._on_response.connect(callback, weak=autoremove)

    def on_error(self, callback, autoremove=True):
        """Register `callback` to receive request exceptions (see `on_response`)"""
        log.debug('Registering %r to receive %s errors', _funcname(callback), self._request_str)
        self._on_error.connect(callback, weak=autoremove)

    @property
    def has_callbacks(self):
        """Whether anyone is interested in response to callback"""
        return bool(self._on_response.receivers) or \
               bool(self._on_error.receivers)

    @property
    def interval(self):
        """Seconds between polls"""
        return self._interval

    @interval.setter
    def interval(self, interval):
        self._interval = interval
        if self.running:
            self.poll()

    def __repr__(self):
        return '<{} {}>'.format(type(self).__name__, self._request_str)



# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details
# http://www.gnu.org/licenses/gpl-3.0.txt

"""TUI and CLI specs for torrent list columns"""

from ..logging import make_logger
log = make_logger(__name__)

from . import ColumnBase


COLUMNS = {}

class Filename(ColumnBase):
    header = {'left': 'Filename'}
    align = 'left'
    width = None
    may_have_wide_chars = True

    def get_value(self):
        return self.data['name']

COLUMNS['name'] = Filename


class Size(ColumnBase):
    header = {'left': 'Size', 'right': '?'}
    width = 6

    def get_value(self):
        return self.data['size-total']

    def get_raw(self):
        return int(self.get_value())

    @classmethod
    def set_unit(cls, unit):
        cls.header['right'] = unit

COLUMNS['size'] = Size


class Downloaded(ColumnBase):
    header = {'left': 'Dn', 'right': '?'}
    width = 6

    def get_value(self):
        return self.data['size-downloaded']

    def get_raw(self):
        return int(self.get_value())

    @classmethod
    def set_unit(cls, unit):
        cls.header['right'] = unit

COLUMNS['downloaded'] = Downloaded


class Progress(ColumnBase):
    header = {'right': '%'}
    width = 4

    def get_value(self):
        return self.data['progress']

    def get_raw(self):
        return int(self.get_value())

COLUMNS['progress'] = Progress


class Priority(ColumnBase):
    header = {'left': 'Priority'}
    width = 4
    align = 'left'

    def get_value(self):
        val = self.get_raw()
        return '' if val == 'normal' else val

    def get_raw(self):
        return 'shun' if self.data['is-wanted'] is False else self.data['priority']

COLUMNS['priority'] = Priority


class Marked(ColumnBase):
    interfaces = ('tui',)

COLUMNS['marked'] = Marked



class TorrentFileDirectory(dict):
    nodetype = 'parent'

    def __hash__(self):
        return hash(self['path'])

    def __repr__(self):
        return '<{} {!r}>'.format(type(self).__name__, self['path'])

def create_directory_data(name, tree, filtered_count=0):
    # Create a mapping that has the same keys as a TorrentFile instance.
    # Each value recursively summarizes the values of all the TorrentFiles
    # in `tree`.

    tfiles = tuple(tree.files)

    def sum_size(tfiles, key):
        sizes = tuple(tfile[key] for tfile in tfiles)
        # Preserve the original type (Number)
        first_size = sizes[0]
        start_value = type(first_size)(0, unit=first_size.unit, prefix=first_size.prefix)
        return sum(sizes, start_value)

    def sum_priority(tfiles):
        if len(set(tfile['priority'] for tfile in tfiles)) == 1:
            return tfiles[0]['priority']
        else:
            return ''

    data = {'size-downloaded': sum_size(tfiles, 'size-downloaded'),
            'size-total': sum_size(tfiles, 'size-total'),
            'priority': sum_priority(tfiles),
            'is-wanted': True}

    data['name'] = create_directory_name(name, filtered_count)

    progress_cls = type(tfiles[0]['progress'])
    data['progress'] = progress_cls(data['size-downloaded'] / data['size-total'] * 100)
    data['tid'] = tfiles[0]['tid']
    data['path'] = tree.path
    data['id'] = frozenset(tf['id'] for tf in tfiles)
    return TorrentFileDirectory(data)

def create_directory_name(name, filtered_count):
    if filtered_count > 0:
        return '%s (%d file%s filtered)' % (name, filtered_count,
                                            '' if filtered_count == 1 else 's')
    else:
        return str(name)



# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details
# http://www.gnu.org/licenses/gpl-3.0.txt

from ...logging import make_logger
log = make_logger(__name__)

from . import (SortSpecBase, SorterBase)

class _SortSpec(SortSpecBase):
    def __init__(self, *args, description='', **kwargs):
        description = 'Sort peers by %s' % description
        super().__init__(*args, description=description, **kwargs)


class TorrentPeerSorter(SorterBase):
    SORTSPECS = {
        'ip':         _SortSpec(lambda t: t['ip'],
                                description='IP address (alphabetically)'),
        'port':       _SortSpec(lambda t: t['port'],
                                description='port number'),
        'client':     _SortSpec(lambda t: t['client'].lower(),
                                description='client name'),
        'country':     _SortSpec(lambda t: t['country'].lower(),
                                 description='country'),
        'progress':   _SortSpec(lambda t: t['progress'],
                                description='downloading progress'),
        'rate-up':    _SortSpec(lambda t: t['rate-up'],
                                description='upload rate (from your perspective)'),
        'rate-down':  _SortSpec(lambda t: t['rate-down'],
                                description='download rate (from your perspective)'),
        'rate':       _SortSpec(lambda t: t['rate-up'] + t['rate-down'],
                                description='combined download and upload rate'),
        'eta':        _SortSpec(lambda t: t['eta'],
                                description='estimated time they need to finish'),
        'rate-est':   _SortSpec(lambda t: t['rate-est'],
                                description='estimated overall download rate'),
        'name':       _SortSpec(lambda t: t['tname'].lower(),
                                description='torrent name'),
    }
    DEFAULT_SORT = 'ip'



# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details
# http://www.gnu.org/licenses/gpl-3.0.txt

from .errors import *
from .constants import *

from .utils import Response

from .sorters.tsorter import TorrentSorter
from .sorters.psorter import TorrentPeerSorter

from .filters.tfilter import TorrentFilter
from .filters.ffilter import TorrentFileFilter
from .filters.pfilter import TorrentPeerFilter

from .poll import RequestPoller
from .trequestpool import TorrentRequestPool

from .aiotransmission.rpc import TransmissionRPC
from .aiotransmission.api_status import StatusAPI
from .aiotransmission.api_settings import SettingsAPI
from .aiotransmission.api_torrent import TorrentAPI
from .api import API



# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details
# http://www.gnu.org/licenses/gpl-3.0.txt

from ...logging import make_logger
log = make_logger(__name__)

import blinker
from types import SimpleNamespace

from ..poll import RequestPoller
from .. import convert
from .. import constants as const


class StatusAPI():
    """Transmission daemon status information"""

    # Pass poller methods through to our pollers
    async def start(self, *args, **kwargs):
        await self._poller_stats.start(*args, **kwargs)
        await self._poller_tcount.start(*args, **kwargs)

    async def stop(self, *args, **kwargs):
        await self._poller_stats.stop(*args, **kwargs)
        await self._poller_tcount.stop(*args, **kwargs)

    def poll(self, *args, **kwargs):
        self._poller_stats.poll(*args, **kwargs)
        self._poller_tcount.poll(*args, **kwargs)

    @property
    def running(self):
        return self._poller_stats.running

    @property
    def interval(self):
        return self._poller_stats.interval

    @interval.setter
    def interval(self, interval):
        self._poller_stats.interval = interval
        self._poller_tcount.interval = interval


    def __init__(self, srvapi, interval=1):
        self._session_stats_updated = False
        self._tcounts_updated = False
        self._reset_session_stats()
        self._reset_tcounts()
        self._on_update = blinker.Signal()

        self._poller_stats = RequestPoller(srvapi.rpc.session_stats,
                                           autoconnect=False,
                                           interval=interval,
                                           loop=srvapi.loop)
        self._poller_stats.on_response(self._handle_session_stats)
        self._poller_stats.on_error(lambda e: log.debug('Ignoring exception: %r', e),
                                    autoremove=False)

        # 'session-stats' provides some counters, but not enough, so we
        # request a minimalistic torrent list.
        self._poller_tcount = RequestPoller(srvapi.torrent.torrents,
                                            keys=('rate-down', 'rate-up', 'isolated'),
                                            autoconnect=False,
                                            interval=interval,
                                            loop=srvapi.loop)
        self._poller_tcount.on_response(self._handle_tlist)

    def _reset_session_stats(self):
        self._session_stats = {}

    def _reset_tcounts(self):
        self._tcounts = SimpleNamespace(
            active=const.DISCONNECTED, downloading=const.DISCONNECTED, isolated=const.DISCONNECTED,
            stopped=const.DISCONNECTED, total=const.DISCONNECTED, uploading=const.DISCONNECTED)

    def _handle_session_stats(self, stats):
        if stats is None:
            self._reset_session_stats()
        else:
            self._session_stats = stats
            self._tcounts.total = stats['torrentCount']
            self._tcounts.stopped = stats['pausedTorrentCount']
            self._tcounts.active = stats['activeTorrentCount']
        self._session_stats_updated = True
        self._maybe_run_callbacks()

    def _handle_tlist(self, response):
        if response is None:
            self._reset_tcounts()
        else:
            tlist = response.torrents
            self._tcounts.isolated = len(tuple(filter(
                lambda t: t['isolated'],
                tlist)))
            self._tcounts.downloading = len(tuple(filter(
                lambda t: t['rate-down'] > 0,
                tlist)))
            self._tcounts.uploading = len(tuple(filter(
                lambda t: t['rate-up'] > 0,
                tlist)))
        self._tcounts_updated = True
        self._maybe_run_callbacks()

    def _maybe_run_callbacks(self):
        # We have two pollers, but we want to call callbacks once when both
        # have an update to report.
        if self._tcounts_updated and self._session_stats_updated:
            self._on_update.send(self)
            self._tcounts_updated = False
            self._session_stats_updated = False

    def on_update(self, callback, autoremove=True):
        """Register `callback` to be called at intervals

        `callback` is passed the instance of this class.  If `autoremove` is
        True, `callback` is stored as a weakref and removed automatically when
        it is deleted.
        """
        log.debug('Registering %r to receive status updates', callback)
        self._on_update.connect(callback, weak=autoremove)

    @property
    def count(self):
        """Torrent counts by category"""
        return self._tcounts

    @property
    def rate_down(self):
        """Total download rate or `constants.DISCONNECTED`"""
        if 'downloadSpeed' in self._session_stats:
            return convert.bandwidth(self._session_stats['downloadSpeed'], unit='byte')
        else:
            return const.DISCONNECTED

    @property
    def rate_up(self):
        """Total upload rate or `constants.DISCONNECTED`"""
        if 'downloadSpeed' in self._session_stats:
            return convert.bandwidth(self._session_stats['uploadSpeed'], unit='byte')
        else:
            return const.DISCONNECTED



# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details
# http://www.gnu.org/licenses/gpl-3.0.txt

"""Base classes for documentation commands"""

from ...logging import make_logger
log = make_logger(__name__)

from .. import (InitCommand, ExpectedResource)
from ... import (APPNAME, __version__)

TOPIC_DELIMITER = ['', '='*50, '']

class HelpCmdbase(metaclass=InitCommand):
    name = 'help'
    aliases = ('man',)
    category = 'misc'
    provides = set()
    description = 'List or explain commands and settings'
    usage = ('help [<TOPIC> <TOPIC> <TOPIC> ...]',)
    examples = ('help',
                'help help')
    argspecs = (
        { 'names': ('TOPIC',), 'nargs': '*',
          'description': 'Name of command, setting or category' },
    )
    helpmgr = ExpectedResource

    def run(self, TOPIC):
        topics = TOPIC
        lines = []
        success = True
        existing_topics = []

        if len(topics) < 1:
            lines = self.helpmgr.overview
        else:
            for topic in topics:
                try:
                    l = self.helpmgr.find(topic)
                except ValueError as e:
                    log.error(e)
                    success = False
                else:
                    lines.extend(l)
                    lines.extend(TOPIC_DELIMITER)
                    existing_topics.append(topic)

            if lines:
                # Remove last TOPIC_DELIMITER
                for _ in range(len(TOPIC_DELIMITER)):
                    lines.pop(-1)

        if not existing_topics:
            existing_topics.append(APPNAME)

        self.display_help(existing_topics, lines)
        return success


class VersionCmdbase(metaclass=InitCommand):
    name = 'version'
    category = 'misc'
    provides = set()
    description = 'Show {} version'.format(APPNAME)

    def run(self):
        log.info('{} version {}'.format(APPNAME, __version__))
        return True



# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details
# http://www.gnu.org/licenses/gpl-3.0.txt

"""Filtering Torrents by their values"""

from ...logging import make_logger
log = make_logger(__name__)

import re

from . import (BoolFilterSpec, CmpFilterSpec, Filter, FilterChain)

from ..tkeys import TYPES as VALUETYPES
def _make_cmp_filter(key, aliases, description):
    filterfunc = lambda t, op, v, key=key: op(t[key], v)
    return CmpFilterSpec(filterfunc,
                         description=_make_filter_desc(description),
                         needed_keys=(key,), aliases=aliases,
                         value_type=VALUETYPES[key])

def _make_filter_desc(text):
    if text.startswith('...'):
        text = 'Match VALUE against ' + text[4:]
    return text



class SingleTorrentFilter(Filter):
    DEFAULT_FILTER = 'name'

    # Filters without arguments
    BOOLEAN_FILTERS = {
        # '...' is replaced with 'Torrents that are'
        'active': BoolFilterSpec(
            lambda t: t['peers-connected'] > 0 or t['status'] == 'verifying',
            description='Torrents connected to peers or being verified',
            needed_keys=('peers-connected', 'status')),
        'all': BoolFilterSpec(
            lambda t: True,
            description='All torrents',
            needed_keys=(),
            aliases=('*',)),
        'complete': BoolFilterSpec(
            lambda t: t['%downloaded'] >= 100,
            description='Torrents with all wanted files complete',
            needed_keys=('%downloaded',)),
        'downloading': BoolFilterSpec(
            lambda t: t['rate-down'] > 0,
            description='Torrents using download bandwidth',
            needed_keys=('rate-down',)),
        'idle': BoolFilterSpec(
            lambda t: t['stalled'],
            description='Torrents not down- or uploading but not stopped',
            needed_keys=('stalled',)),
        'isolated': BoolFilterSpec(
            lambda t: t['isolated'],
            description='Torrents that cannot discover new peers in any way',
            needed_keys=('isolated',)),
        'leeching': BoolFilterSpec(
            lambda t: t['status'] == 'leeching',
            description='Torrents downloading or waiting for seeds',
            needed_keys=('status',)),
        'private': BoolFilterSpec(
            lambda t: t['private'],
            description='Torrents that are only connectable through trackers',
            needed_keys=('private',)),
        'public': BoolFilterSpec(
            lambda t: not t['private'],
            description='Torrents connectable through DHT and/or PEX',
            needed_keys=('private',)),
        'seeding': BoolFilterSpec(
            lambda t: t['status'] == 'seeding',
            description='Torrents that are complete and shared on request',
            needed_keys=('status',)),
        'stopped': BoolFilterSpec(
            lambda t: t['status'] == 'stopped',
            description='Torrents not allowed to up- or download',
            needed_keys=('status',),
            aliases=('paused',)),
        'uploading': BoolFilterSpec(
            lambda t: t['rate-up'] > 0,
            description='Torrents using upload bandwidth',
            needed_keys=('rate-up',)),
        'verifying': BoolFilterSpec(
            lambda t: t['status'] == 'verifying',
            description='Torrents being verified or queued for verification',
            needed_keys=('status',),
            aliases=('checking',)),
    }


    # Filters with arguments
    COMPARATIVE_FILTERS = {
        'connections': _make_cmp_filter('peers-connected', ('conn',),
                                        '... number of connected peers'),
        '%downloaded': _make_cmp_filter('%downloaded', ('%down',),
                                        '... percentage of downloaded bytes'),
        'downloaded': _make_cmp_filter('size-downloaded', ('down',),
                                       '... number of downloaded bytes'),
        'id':        _make_cmp_filter('id', (), '... ID'),
        'name':      _make_cmp_filter('name', ('title',), '... name'),
        'path':      _make_cmp_filter('path', ('dir',), '... full path to download directory'),
        'ratio':     _make_cmp_filter('ratio', (), '... uploaded/downloaded ratio'),
        'rate-down': _make_cmp_filter('rate-down', ('rdown',), '... download rate'),
        'rate-up':   _make_cmp_filter('rate-up', ('rup',), '... upload rate'),
        'seeds':     _make_cmp_filter('peers-seeding', (),
                                      '... largest number of seeds reported by any tracker'),
        'size':      _make_cmp_filter('size-final', (),
                                      '... combined size of all wanted files'),
        'uploaded':  _make_cmp_filter('size-uploaded', (),
                                      '... number of uploaded bytes'),

        'tracker': CmpFilterSpec(
            lambda t, op, v: any(op(tracker['url-announce'].domain, v)
                                 for tracker in t['trackers']),
            description=_make_filter_desc('... domain of the announce URL of trackers'),
            needed_keys=('trackers',),
            value_type=str,
        ),
    }



class TorrentFilter(FilterChain):
    """One or more filters combined with & and | operators"""
    filterclass = SingleTorrentFilter



# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details
# http://www.gnu.org/licenses/gpl-3.0.txt

"""Mixin classes that are common between TUI and CLI commands"""

from ...logging import make_logger
log = make_logger(__name__)


from .. import utils


class get_torrent_sorter():
    def get_torrent_sorter(self, args):
        """Return TorrentSorter instance or None

        If `args` evaluates to True, it is passed to TorrentSorter and the
        result is returned.

        If `args` evaluates to False, None is returned.
        """
        if args:
            from ...client import TorrentSorter
            return TorrentSorter(utils.listify_args(args))


class get_peer_sorter():
    def get_peer_sorter(self, args):
        """Return TorrentPeerSorter instance or None

        If `args` evaluates to True, it is passed to TorrentPeerSorter and the
        result is returned.

        If `args` evaluates to False, None is returned.
        """
        if args:
            from ...client import TorrentPeerSorter
            return TorrentPeerSorter(utils.listify_args(args))


class get_peer_filter():
    def get_peer_filter(self, FILTER):
        """Return TorrentPeerFilter instance or None

        If `FILTER` evaluates to True, it is passed to TorrentPeerFilter and
        the resulting object is returned.

        If `FILTER` evaluates to False, None is returned.
        """
        if FILTER:
            from ...client import TorrentPeerFilter
            return TorrentPeerFilter(FILTER)


class get_tlist_columns():
    def get_tlist_columns(self, columns, interface=None):
        """Check if each item in iterable `columns` is a valid torrent list column name

        If `interface` is not None, also remove all columns that don't have
        `interface` in their `interfaces` attribute.

        Raise ValueError or return a new list of `columns`.
        """
        from ...columns.tlist import COLUMNS
        cols = utils.listify_args(columns)
        for col in tuple(cols):
            if col not in COLUMNS:
                raise ValueError('Unknown column name: {}'.format(col))
            elif interface is not None and interface not in COLUMNS[col].interfaces:
                log.debug('Removing column %r because it does not support %r',
                          col, interface)
                cols.remove(col)
        return cols


class get_flist_columns():
    def get_flist_columns(self, columns):
        """Check if each item in iterable `columns` is a valid file list column name

        Raise ValueError or return a new list of `columns`.
        """
        from ...columns.flist import COLUMNS
        cols = utils.listify_args(columns)
        for col in cols:
            if col not in COLUMNS:
                raise ValueError('Unknown column name: {}'.format(col))
        return cols


class get_plist_columns():
    def get_plist_columns(self, columns):
        """Check if each item in iterable `columns` is a valid peer list column name

        Raise ValueError or return a new list of `columns`.
        """
        from ...columns.plist import COLUMNS
        cols = utils.listify_args(columns)
        for col in cols:
            if col not in COLUMNS:
                raise ValueError('Unknown column name: {}'.format(col))
        return cols
